# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI VDIF payloads.

Implements a VDIFPayload class used to store payload words, and decode to
or encode from a data array.

See the `VDIF specification page <http://www.vlbi.org/vdif>`_ for payload
specifications.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np
from collections import namedtuple

from ..vlbi_base.payload import VLBIPayloadBase
from ..vlbi_base.encoding import (encode_2bit_base, encode_4bit_base,
                                  decoder_levels, decode_8bit, encode_8bit)

__all__ = ['init_luts', 'decode_2bit', 'decode_4bit', 'encode_2bit',
           'encode_4bit', 'VDIFPayload']


def init_luts():
    """Sets up the look-up tables for levels as a function of input byte.

    Returns
    -------
    lut1bit : `~numpy.ndarray`
        Look-up table for decoding bytes to 1-bit samples.
    lut2bit : `~numpy.ndarray`
        As `lut1bit1`, but for 2-bit samples.
    lut4bit : `~numpy.ndarray`
        As `lut1bit1`, but for 4-bit samples.

    Notes
    -----

    Look-up tables are two-dimensional arrays whose first axis is indexed
    by byte value (in uint8 form) and whose second axis represents sample
    temporal order.  Table values are decoded sample values.  Sec. 10 in
    the `VDIF Specification
    <http://vlbi.org/vdif/docs/VDIF_specification_Release_1.1.1.pdf>`_
    states that samples are encoded by offset-binary, such that all 0
    bits is lowest and all 1 bits is highest.  I.e., for 2-bit sampling,
    the order is 00, 01, 10, 11.  These are decoded using
    `~baseband.vlbi_base.encoding.decoder_levels`.

    For example, the 2-bit sample sequence ``-1, -1, 1, 1`` is encoded
    as ``0b10100101`` (or ``165`` in uint8 form).  To translate this back
    to sample values, access ``lut2bit`` using the byte as the key::

        >>> lut2bit[0b10100101]
        array([-1., -1.,  1.,  1.], dtype=float32)
    """
    b = np.arange(256)[:, np.newaxis]
    # 1-bit mode
    i = np.arange(8)
    lut1bit = decoder_levels[1][(b >> i) & 1]
    # 2-bit mode
    i = np.arange(0, 8, 2)
    lut2bit = decoder_levels[2][(b >> i) & 3]
    # 4-bit mode
    i = np.arange(0, 8, 4)
    lut4bit = decoder_levels[4][(b >> i) & 0xf]
    return lut1bit, lut2bit, lut4bit


lut1bit, lut2bit, lut4bit = init_luts()


def decode_2bit(words):
    """Decodes data stored using 2 bits per sample."""
    b = words.view(np.uint8)
    return lut2bit.take(b, axis=0)


shift2bit = np.arange(0, 8, 2).astype(np.uint8)


def encode_2bit(values):
    """Encodes values using 2 bits per sample, packing the result into bytes.
    """
    bitvalues = encode_2bit_base(values.reshape(-1, 4))
    bitvalues <<= shift2bit
    return np.bitwise_or.reduce(bitvalues, axis=-1)


def decode_4bit(words):
    """Decodes data stored using 4 bits per sample."""
    b = words.view(np.uint8)
    return lut4bit.take(b, axis=0)


shift04 = np.array([0, 4], np.uint8)


def encode_4bit(values):
    """Encodes values using 4 bits per sample, packing the result into bytes.
    """
    b = encode_4bit_base(values).reshape(-1, 2)
    b <<= shift04
    return b[:, 0] | b[:, 1]


class VDIFPayload(VLBIPayloadBase):
    """Container for decoding and encoding VDIF payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    header : `~baseband.vdif.VDIFHeader`
        If given, used to infer the number of channels, bps, and whether
        the data are complex.
    nchan : int, optional
        Number of channels, used if ``header`` is not given.  Default: 1.
    bps : int, optional
        Bits per elementary sample, used if ``header`` is not given.
        Default: 2.
    complex_data : bool, optional
        Whether the data are complex, used if ``header`` is not given.
        Default: `False`.
    """
    _decoders = {2: decode_2bit,
                 4: decode_4bit,
                 8: decode_8bit}

    _encoders = {2: encode_2bit,
                 4: encode_4bit,
                 8: encode_8bit}

    _sample_shape_maker = namedtuple('SampleShape', 'nchan')

    def __init__(self, words, header=None, nchan=1, bps=2, complex_data=False):
        if header is not None:
            nchan = header.nchan
            bps = header.bps
            complex_data = header['complex_data']
            self._nbytes = header.payload_nbytes
            if header.edv == 0xab:  # Mark5B payload
                from ..mark5b import Mark5BPayload
                self._decoders = Mark5BPayload._decoders
                self._encoders = Mark5BPayload._encoders
                if complex_data:
                    raise ValueError("VDIF/Mark5B payload cannot be complex.")
        super(VDIFPayload, self).__init__(words, sample_shape=(nchan,),
                                          bps=bps, complex_data=complex_data)

    @classmethod
    def fromfile(cls, fh, header):
        """Read payload from filehandle and decode it into data.

        Parameters
        ----------
        fh : filehandle
            To read data from.
        header : `~baseband.vdif.VDIFHeader`
            Used to infer the payload size, number of channels, bits per
            sample, and whether the data are complex.
        """
        s = fh.read(header.payload_nbytes)
        if len(s) < header.payload_nbytes:
            raise EOFError("could not read full payload.")
        return cls(np.frombuffer(s, dtype=cls._dtype_word), header)

    @classmethod
    def fromdata(cls, data, header=None, bps=2, edv=None):
        """Encode data as payload, using header information.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Values to be encoded.
        header : `~baseband.vdif.VDIFHeader`, optional
            If given, used to infer the encoding, and to verify the number of
            channels and whether the data are complex.
        bps : int, optional
            Bits per elementary sample, used if ``header`` is not given.
            Default: 2.
        edv : int, optional
            Should be given if ``header`` is not given and the payload is
            encoded as Mark 5 data (i.e., ``edv=0xab``).
        """
        nchan = data.shape[-1]
        complex_data = (data.dtype.kind == 'c')
        if header is not None:
            if header.nchan != nchan:
                raise ValueError("header is for {0} channels but data has {1}"
                                 .format(header.nchan, data.shape[-1]))
            if header['complex_data'] != complex_data:
                raise ValueError("header is for {0} data but data are {1}"
                                 .format(*(('complex' if c else 'real') for c
                                           in (header['complex_data'],
                                               complex_data))))
            bps = header.bps
            edv = header.edv

        if edv == 0xab:  # Mark5B payload
            from ..mark5b import Mark5BPayload
            encoder = Mark5BPayload._encoders[bps]
        else:
            encoder = cls._encoders[bps]

        if complex_data:
            data = data.view((data.real.dtype, (2,)))
        words = encoder(data).ravel().view(cls._dtype_word)
        return cls(words, header, nchan=nchan, bps=bps,
                   complex_data=complex_data)



# Licensed under the GPLv3 - see LICENSE
"""Encoders and decoders for generic VLBI data formats."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np


__all__ = ['OPTIMAL_2BIT_HIGH', 'TWO_BIT_1_SIGMA', 'FOUR_BIT_1_SIGMA',
           'EIGHT_BIT_1_SIGMA', 'decoder_levels', 'encode_2bit_base',
           'encode_4bit_base', 'decode_8bit', 'encode_8bit']


# The high mag value for 2-bit reconstruction.  Note that mark5access uses
# OPTIMAL_2BIT_HIGH = 3.3359, which is possibly a typo.
OPTIMAL_2BIT_HIGH = 3.316505
r"""Optimal high value for a 2-bit digitizer for which the low value is 1.

It is chosen such that for a normal distribution in which 68.269% of all
values are at the low level, this is the mean of the others, i.e.,

.. math::

    l = \frac{\int_\sigma^\infty x \exp(-\frac{x^2}{2\sigma^2}) dx}
             {\int_\sigma^\infty \exp(-\frac{x^2}{2\sigma^2}) dx},

where the standard deviation is determined from:

.. math::

    1 = \frac{\int_0^\sigma x \exp(-\frac{x^2}{2\sigma^2}) dx}
         {\int_0^\sigma \exp(-\frac{x^2}{2\sigma^2}) dx}.

These give:

.. math::

    \sigma = \frac{\sqrt{\frac{\pi}{2}}\mathrm{erf}
             (\sqrt{1/2})}{1 - \sqrt{1/e}} = 2.174564,

and

.. math::

    l = \frac{1}{(\sqrt{e} - 1)(1/\mathrm{erf}(\sqrt{1/2}) - 1)} = 3.316505
"""
TWO_BIT_1_SIGMA = 2.174564
"""Optimal level between low and high for the above OPTIMAL_2BIT_HIGH."""
FOUR_BIT_1_SIGMA = 2.95
"""Scaling for four-bit encoding that makes it look like 2 bit."""
EIGHT_BIT_1_SIGMA = 71.0 / 2.
"""Scaling for eight-bit encoding that makes it look like 2 bit."""

decoder_levels = {
    1: np.array([-1.0, 1.0], dtype=np.float32),
    2: np.array([-OPTIMAL_2BIT_HIGH, -1.0, 1.0, OPTIMAL_2BIT_HIGH],
                dtype=np.float32),
    4: (np.arange(16, dtype=np.float32) - 8.) / FOUR_BIT_1_SIGMA}
"""Levels for data encoded with different numbers of bits.."""

two_bit_2_sigma = 2 * TWO_BIT_1_SIGMA
clip_low, clip_high = -1.5 * TWO_BIT_1_SIGMA, 1.5 * TWO_BIT_1_SIGMA


def encode_2bit_base(values):
    """Generic encoder for data stored using two bits.

    This returns an unsigned integer array containing encoded sample
    values that range from 0 to 3.  The conversion from floating point
    sample value to unsigned int is given below, with
    ``lv = TWO_BIT_1_SIGMA = 2.1745``:

      ================= ======
      Input range       Output
      ================= ======
            value < -lv   0
      -lv < value <  0.   2
       0. < value <  lv   1
       lv < value         3
      ================= ======

    This does not pack the samples into bytes.
    """
    # Optimized for speed by doing calculations in-place, and ensuring that
    # the dtypes match.
    values = np.clip(values, clip_low, clip_high)
    values += two_bit_2_sigma
    bitvalues = np.empty(values.shape, np.uint8)
    return np.floor_divide(values, TWO_BIT_1_SIGMA, out=bitvalues,
                           casting='unsafe')


def encode_4bit_base(values):
    """Generic encoder for data stored using four bits.

    This returns an unsigned integer array containing encoded sample
    values that range from 0 to 15.  Floating point sample values are
    converted to unsigned int by first scaling them by
    ``FOUR_BIT_1_SIGMA = 2.95``, then adding 8.5 (the 0.5 to ensure proper
    rounding when typecasting to uint8). Some sample output levels are:

      ========================= ======
      Input range               Output
      ========================= ======
             value*scale < -7.5    0
      -7.5 < value*scale < -6.5    1
      -0.5 < value*scale < +0.5    8
       6.5 < value*scale          15
      ========================= ======

    This does not pack the samples into bytes.
    """
    # Optimized for speed by doing calculations in-place.
    values = values * FOUR_BIT_1_SIGMA
    values += 8.5
    return np.clip(values, 0., 15., out=values).astype(np.uint8)


def decode_8bit(words):
    """Generic decoder for data stored using 8 bits.

    We follow mark5access, which assumes the values 0 to 255 encode
    -127.5 to 127.5, scaled down to match 2 bit data by a factor of 35.5
    (`~baseband.vlbi_base.encoding.EIGHT_BIT_1_SIGMA`)

    For comparison, GMRT phased data treats the 8-bit data values simply
    as signed integers.
    """
    b = words.view(np.uint8).astype(np.float32)
    b -= 127.5
    b /= EIGHT_BIT_1_SIGMA
    return b


def encode_8bit(values):
    """Encode 8 bit VDIF data.

    We follow mark5access, which assumes the values 0 to 255 encode
    -127.5 to 127.5, scaled down to match 2 bit data by a factor of 35.5
    (`~baseband.vlbi_base.encoding.EIGHT_BIT_1_SIGMA`)

    For comparison, GMRT phased data treats the 8-bit data values simply
    as signed integers.
    """
    return (np.clip(np.rint(values * EIGHT_BIT_1_SIGMA + 127.5), 0, 255)
            .astype(np.uint8))



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..vlbi_base.file_info import VLBIFileReaderInfo


class VDIFFileReaderInfo(VLBIFileReaderInfo):
    attr_names = ('format', 'edv') + VLBIFileReaderInfo.attr_names[1:]
    _header0_attrs = ('edv', 'bps', 'samples_per_frame')

    def _get_header0(self):
        fh = self._parent
        old_offset = fh.tell()
        try:
            fh.seek(0)
            header0 = fh.read_header()
            # Almost all bytes are interpretable as headers,
            # so we need a basic sanity check.
            fh.seek(header0.frame_nbytes)
            header1 = fh.read_header()
            if header1.same_stream(header0):
                return header0
        except Exception:
            pass
        finally:
            fh.seek(old_offset)

    def _get_start_time(self):
        try:
            return self.header0.get_time(frame_rate=self.frame_rate)
        except Exception:
            return None

    def _get_sample_shape(self):
        # To get the sample shape, need to read a while frameset.
        fh = self._parent
        old_offset = fh.tell()
        try:
            fh.seek(0)
            return fh.read_frameset().sample_shape
        except Exception:
            return None
        finally:
            fh.seek(old_offset)

    def _collect_info(self):
        super(VDIFFileReaderInfo, self)._collect_info()
        if self:
            self.complex_data = self.header0['complex_data']
            self.sample_shape = self._get_sample_shape()



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import pytest
import numpy as np
import astropy.units as u
from astropy.time import Time
from .. import vdif
from .. import mark4
from .. import mark5b
from .. import dada
from ..vlbi_base.encoding import EIGHT_BIT_1_SIGMA
from ..data import (SAMPLE_MARK4 as SAMPLE_M4, SAMPLE_MARK5B as SAMPLE_M5B,
                    SAMPLE_VDIF, SAMPLE_MWA_VDIF as SAMPLE_MWA, SAMPLE_DADA)


class TestVDIFMark5B(object):
    """Simplest conversion: VDIF frame containing Mark5B data (EDV 0xab)."""

    def test_header(self):
        """Check Mark 5B header information can be stored in a VDIF header."""
        with open(SAMPLE_M5B, 'rb') as fh:
            # Start time kiloday is needed for Mark 5B to calculate time.
            m5h1 = mark5b.Mark5BHeader.fromfile(fh, kday=56000)
            # For the payload, pass in how data is encoded.
            m5pl = mark5b.Mark5BPayload.fromfile(fh, nchan=8, bps=2)
            # A not-at-the-start header for checking times.
            m5h2 = mark5b.Mark5BHeader.fromfile(fh, kday=56000)
        # Create VDIF headers based on both the Mark 5B header and payload.
        header1 = vdif.VDIFHeader.from_mark5b_header(
            m5h1, nchan=m5pl.sample_shape.nchan, bps=m5pl.bps)
        header2 = vdif.VDIFHeader.from_mark5b_header(
            m5h2, nchan=m5pl.sample_shape.nchan, bps=m5pl.bps)
        for i, (m5h, header) in enumerate(((m5h1, header1), (m5h2, header2))):
            assert m5h['frame_nr'] == i
            # Check all direct information is set correctly.
            assert all(m5h[key] == header[key] for key in m5h.keys())
            assert header['mark5b_frame_nr'] == m5h['frame_nr']
            assert header.kday == m5h.kday
            # As well as the time calculated from the header information.
            assert header.time == m5h.time
            # Check information on the payload is also correct.
            assert header.nchan == 8
            assert header.bps == 2
            assert not header['complex_data']
            assert header.frame_nbytes == 10032
            assert header.nbytes == 32
            assert header.payload_nbytes == m5h.payload_nbytes
            assert (header.samples_per_frame ==
                    10000 * 8 // m5pl.bps // m5pl.sample_shape.nchan)

        # Check that we can handle > 512 Mbps sampling rate.
        header3 = vdif.VDIFHeader.from_mark5b_header(
            m5h2, nchan=m5pl.sample_shape.nchan, bps=m5pl.bps,
            sample_rate=64*u.MHz)
        assert header3.time == header2.time
        assert header3['frame_nr'] == m5h2['frame_nr']

        # A copy might remove any `kday` keywords set, but should still work
        # (Regression test for #34)
        header_copy = header2.copy()
        assert header_copy == header2
        header_copy.verify()
        # But it should not remove `kday` to start with (#35)
        assert header_copy.kday == header2.kday
        # May as well check that with a corrupt 'bcd_fraction' we can still
        # get the right time using the frame number.
        header_copy['bcd_fraction'] = 0
        # This is common enough that we should not fail verification.
        header_copy.verify()
        # However, it should also cause just getting the time to fail
        # unless we pass in a frame rate.
        with pytest.raises(ValueError):
            header_copy.time
        frame_rate = 32. * u.MHz / header.samples_per_frame
        assert abs(header_copy.get_time(frame_rate=frame_rate) -
                   m5h2.time) < 1.*u.ns

    def test_payload(self):
        """Check Mark 5B payloads can used in a Mark5B VDIF payload."""
        # Get Mark 5B header, payload, and construct VDIF header, as above.
        with open(SAMPLE_M5B, 'rb') as fh:
            m5h = mark5b.Mark5BHeader.fromfile(fh, kday=56000)
            m5pl = mark5b.Mark5BPayload.fromfile(fh, nchan=8, bps=2)
        header = vdif.VDIFHeader.from_mark5b_header(
            m5h, nchan=m5pl.sample_shape.nchan, bps=m5pl.bps)
        # Create VDIF payload from the Mark 5B encoded payload.
        payload = vdif.VDIFPayload(m5pl.words, header)
        # Check that the payload (i.e., encoded data) is the same.
        assert np.all(payload.words == m5pl.words)
        # And check that if we decode the payload, we get the same result.
        assert np.all(payload.data == m5pl.data)
        # Now construct a VDIF payload from the Mark 5B data, checking that
        # the encoding works correctly too.
        payload2 = vdif.VDIFPayload.fromdata(m5pl.data, header)
        assert np.all(payload2.words == m5pl.words)
        assert np.all(payload2.data == m5pl.data)
        # Mark 5B data cannot complex. Check that this raises an exception.
        header2 = header.copy()
        header2['complex_data'] = True
        with pytest.raises(ValueError):
            vdif.VDIFPayload(m5pl.words, header2)

    def test_frame(self):
        """Check a whole Mark 5B frame can be translated to VDIF."""
        with mark5b.open(SAMPLE_M5B, 'rb', ref_time=Time(57000, format='mjd'),
                         nchan=8, bps=2) as fh:
            # pick second frame just to be different from header checks above.
            fh.seek(10016)
            m5f = fh.read_frame()

        assert m5f['frame_nr'] == 1
        frame = vdif.VDIFFrame.from_mark5b_frame(m5f)
        assert frame.nbytes == 10032
        assert frame.shape == (5000, 8)
        assert np.all(frame.data == m5f.data)
        assert frame.time == m5f.time

    def test_stream(self):
        """Check we can encode a whole stream."""


class TestVDIF0VDIF1(object):
    """Conversion between EDV versions."""

    def test_stream(self, tmpdir):
        with vdif.open(SAMPLE_MWA, 'rs', sample_rate=1.28*u.MHz) as f0:
            h0 = f0.header0
            d0 = f0.read(1024)
            kwargs = dict(h0)
            kwargs['edv'] = 1
            fl = str(tmpdir.join('test1.vdif'))
            with vdif.open(fl, 'ws', sample_rate=1.28*u.MHz, **kwargs) as f1w:
                h1w = f1w.header0
                assert list(h1w.words[:4]) == list(h0.words[:4])
                assert h1w.sample_rate == 1.28*u.MHz
                f1w.write(d0)

            with vdif.open(fl, 'rs') as f1r:
                h1r = f1r.header0
                d1r = f1r.read(1024)
                assert h1r.words[:4] == h0.words[:4]
                assert h1w.sample_rate == 1.28*u.MHz
                assert np.all(d1r == d0)


class TestMark5BToVDIF3(object):
    """Real conversion: Mark5B to VDIF EDV 3, and back to Mark5B"""

    def test_header(self):
        with open(SAMPLE_M5B, 'rb') as fh:
            m5h = mark5b.Mark5BHeader.fromfile(fh, kday=56000)
            m5pl = mark5b.Mark5BPayload.fromfile(fh, nchan=8, bps=2)
        # check that we have enough information to create VDIF EDV 3 header.
        header = vdif.VDIFHeader.fromvalues(
            edv=3, bps=m5pl.bps, nchan=1, station='WB', time=m5h.time,
            sample_rate=32.*u.MHz, complex_data=False)
        assert header.time == m5h.time

    def test_stream(self, tmpdir):
        """Convert Mark 5B data stream to VDIF."""
        # Here, we need to give how the data is encoded, since the data do not
        # self-describe this.  Furthermore, we need to pass in a rough time,
        # and the rate at which samples were taken, so that absolute times can
        # be calculated.
        with mark5b.open(SAMPLE_M5B, 'rs', sample_rate=32.*u.MHz, kday=56000,
                         nchan=8, bps=2) as fr:
            m5h = fr.header0
            # create VDIF header from Mark 5B stream information.
            header = vdif.VDIFHeader.fromvalues(
                edv=3, bps=fr.bps, nchan=1, station='WB', time=m5h.time,
                sample_rate=32.*u.MHz, complex_data=False)
            data = fr.read(20000)  # enough to fill one EDV3 frame.
            time1 = fr.tell(unit='time')

        # Get a file name in our temporary testing directory.
        vdif_file = str(tmpdir.join('converted.vdif'))
        # create and fill vdif file with converted data.
        with vdif.open(vdif_file, 'ws', header0=header,
                       nthread=data.shape[1]) as fw:
            assert (fw.tell(unit='time') - m5h.time) < 2. * u.ns
            fw.write(data)
            assert (fw.tell(unit='time') - time1) < 2. * u.ns

        # Check two files contain same information.
        with mark5b.open(SAMPLE_M5B, 'rs', sample_rate=32.*u.MHz, kday=56000,
                         nchan=8, bps=2) as fm, vdif.open(vdif_file,
                                                          'rs') as fv:
            assert fm.header0.time == fv.header0.time
            dm = fm.read(20000)
            dv = fv.read(20000)
            assert np.all(dm == dv)
            assert fm.offset == fv.offset
            assert fm.tell(unit='time') == fv.tell(unit='time')

            # Convert VDIF file back to Mark 5B
            mark5b_new_file = str(tmpdir.join('reconverted.mark5b'))
            hv = fv.header0
            hm = fm.header0
            # Here, we fill some unimportant Mark 5B header information by
            # hand, so we can compare byte-for-byte.
            with mark5b.open(mark5b_new_file, 'ws', sample_rate=hv.sample_rate,
                             nchan=dv.shape[1], bps=hv.bps,
                             time=hv.time, user=hm['user'],
                             internal_tvg=hm['internal_tvg']) as fw:
                fw.write(dv)

        with open(SAMPLE_M5B, 'rb') as fh_orig, open(mark5b_new_file,
                                                     'rb') as fh_new:
            assert fh_orig.read() == fh_new.read()


class TestVDIF3ToMark5B(object):
    """Real conversion: VDIF EDV 3 to Mark5B."""

    def test_header(self):
        with open(SAMPLE_VDIF, 'rb') as fh:
            vh = vdif.VDIFHeader.fromfile(fh)

        header = mark5b.Mark5BHeader.fromvalues(time=vh.time)
        assert header.time == vh.time

    def test_stream(self, tmpdir):
        with vdif.open(SAMPLE_VDIF, 'rs') as fr:
            vh = fr.header0
            data = fr.read(20000)  # enough to fill two Mark 5B frames.

        fl = str(tmpdir.join('test.m5b'))
        with mark5b.open(fl, 'ws', sample_rate=vh.sample_rate,
                         nchan=data.shape[1], bps=vh.bps, time=vh.time) as fw:
            fw.write(data)

        with vdif.open(SAMPLE_VDIF, 'rs') as fv, mark5b.open(
                fl, 'rs', sample_rate=32.*u.MHz,
                ref_time=Time(57000, format='mjd'), nchan=8, bps=2) as fm:
            assert fv.header0.time == fm.header0.time
            dv = fv.read(20000)
            dm = fm.read(20000)
            assert np.all(dm == dv)
            assert fm.offset == fv.offset
            assert fm.tell(unit='time') == fv.tell(unit='time')


class TestMark4ToVDIF1(object):
    """Real conversion: Mark 4 to VDIF EDV 1, and back to Mark 4.

    Here, need to use a VDIF format with a flexible size, since we want
    to create invalid frames corresponding to the pieces of data overwritten
    by the Mark 4 header.
    """

    def test_header(self):
        with open(SAMPLE_M4, 'rb') as fh:
            fh.seek(0xa88)
            m4h = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
        # Check that we have enough information to create VDIF EDV 1 header.
        header = vdif.VDIFHeader.fromvalues(
            edv=1, bps=m4h.bps, nchan=1, station='Ar', time=m4h.time,
            sample_rate=32.*u.MHz, payload_nbytes=640*2//8, complex_data=False)
        assert abs(header.time - m4h.time) < 2. * u.ns

    def test_stream(self, tmpdir):
        with mark4.open(SAMPLE_M4, 'rs', sample_rate=32.*u.MHz,
                        ntrack=64, decade=2010) as fr:
            m4header0 = fr.header0
            start_time = fr.start_time
            vheader0 = vdif.VDIFHeader.fromvalues(
                edv=1, bps=m4header0.bps, nchan=1, station='Ar',
                time=start_time, sample_rate=32.*u.MHz,
                payload_nbytes=640*2//8, complex_data=False)
            assert abs(vheader0.time - start_time) < 2. * u.ns
            data = fr.read(80000)  # full Mark 4 frame
            offset1 = fr.tell()
            time1 = fr.tell(unit='time')
            number_of_bytes = fr.fh_raw.tell() - 0xa88

        with open(SAMPLE_M4, 'rb') as fh:
            fh.seek(0xa88)
            orig_bytes = fh.read(number_of_bytes)

        fl = str(tmpdir.join('test.vdif'))
        with vdif.open(fl, 'ws', header0=vheader0,
                       nthread=data.shape[1]) as fw:
            assert (fw.tell(unit='time') - start_time) < 2. * u.ns
            # Write first VDIF frame, matching Mark 4 Header, hence invalid.
            fw.write(data[:160], valid=False)
            # Write remaining VDIF frames, with valid data.
            fw.write(data[160:])
            assert (fw.tell(unit='time') - time1) < 2. * u.ns

        with vdif.open(fl, 'rs') as fv:
            assert abs(fv.header0.time - start_time) < 2. * u.ns
            expected = vheader0.copy()
            expected['invalid_data'] = True
            assert fv.header0 == expected
            dv = fv.read(80000)
            assert np.all(dv == data)
            assert fv.offset == offset1
            assert abs(fv.tell(unit='time') - time1) < 2.*u.ns

        # Convert VDIF file back to Mark 4, and check byte-for-byte.
        fl2 = str(tmpdir.join('test.m4'))
        with mark4.open(fl2, 'ws', sample_rate=vheader0.sample_rate,
                        ntrack=64, bps=2, fanout=4, time=vheader0.time,
                        system_id=108) as fw:
            fw.write(dv)

        with open(fl2, 'rb') as fh:
            conv_bytes = fh.read()
            assert len(conv_bytes) == len(conv_bytes)
            assert orig_bytes == conv_bytes


class TestDADAToVDIF1(object):
    """Real conversion: DADA to VDIF EDV 1, and back to DADA.

    Here, we use a VDIF format with a flexible size so it is easier to fit
    the dada file inside the VDIF one.
    """

    def get_vdif_header(self, header):
        return vdif.VDIFHeader.fromvalues(
            edv=1, time=header.time, sample_rate=header.sample_rate,
            bps=header.bps, nchan=header['NCHAN'],
            complex_data=header.complex_data,
            payload_nbytes=header.payload_nbytes // 2,
            station=header['TELESCOPE'][:2])

    def get_vdif_data(self, dada_data):
        return (dada_data + 0.5 + 0.5j) / EIGHT_BIT_1_SIGMA

    def get_dada_data(self, vdif_data):
        return vdif_data * EIGHT_BIT_1_SIGMA - 0.5 - 0.5j

    def test_header(self):
        with open(SAMPLE_DADA, 'rb') as fh:
            ddh = dada.DADAHeader.fromfile(fh)
        # Check that we have enough information to create VDIF EDV 1 header.
        header = self.get_vdif_header(ddh)
        assert abs(header.time - ddh.time) < 2. * u.ns
        assert header.payload_nbytes == ddh.payload_nbytes // 2

    def test_payload(self):
        with open(SAMPLE_DADA, 'rb') as fh:
            fh.seek(4096)
            ddp = dada.DADAPayload.fromfile(fh, payload_nbytes=64000,
                                            sample_shape=(2, 1),
                                            complex_data=True, bps=8)
        dada_data = ddp.data
        # Check that conversion between scalings works.
        vdif_data = self.get_vdif_data(dada_data)
        assert np.allclose(self.get_dada_data(vdif_data), dada_data)
        # Check that we can create correct payloads.
        vdif_payload0 = vdif.VDIFPayload.fromdata(vdif_data[:, 0, :], bps=8)
        vdif_payload1 = vdif.VDIFPayload.fromdata(vdif_data[:, 1, :], bps=8)
        vd0, vd1 = vdif_payload0.data, vdif_payload1.data
        assert np.allclose(vd0, vdif_data[:, 0, :])
        assert np.allclose(vd1, vdif_data[:, 1, :])
        vd = np.zeros((vd0.shape[0], 2, vd0.shape[1]), vd0.dtype)
        vd[:, 0] = vd0
        vd[:, 1] = vd1
        dd_new = self.get_dada_data(vd)
        ddp2 = dada.DADAPayload.fromdata(dd_new, bps=8)
        assert ddp2 == ddp

    def test_stream(self, tmpdir):
        with dada.open(SAMPLE_DADA, 'rs') as fr:
            ddh = fr.header0
            dada_data = fr.read()
            offset1 = fr.tell()
            stop_time = fr.tell(unit='time')

        header = self.get_vdif_header(ddh)
        data = self.get_vdif_data(dada_data)
        assert abs(header.time - ddh.time) < 2. * u.ns
        vdif_file = str(tmpdir.join('converted_dada.vdif'))
        with vdif.open(vdif_file, 'ws', header0=header,
                       nthread=data.shape[1]) as fw:
            assert (fw.tell(unit='time') - header.time) < 2. * u.ns
            # Write all data in since frameset, made of two frames.
            fw.write(data)
            assert (fw.tell(unit='time') - stop_time) < 2. * u.ns
            assert fw.offset == offset1

        with vdif.open(vdif_file, 'rs') as fv:
            assert abs(fv.header0.time - ddh.time) < 2. * u.ns
            dv = fv.read()
            assert fv.offset == offset1
            assert np.abs(fv.tell(unit='time') - stop_time) < 2.*u.ns
            vh = fv.header0
            vnthread = fv.sample_shape.nthread
        assert np.allclose(dv, data)

        # Convert VDIF file back to DADA.
        dada_file = str(tmpdir.join('reconverted.dada'))
        dv_data = self.get_dada_data(dv)
        assert np.allclose(dv_data, dada_data)
        with dada.open(dada_file, 'ws', sample_rate=vh.sample_rate,
                       time=vh.time, npol=vnthread, bps=vh.bps,
                       payload_nbytes=vh.payload_nbytes*2, nchan=vh.nchan,
                       telescope=vh.station,
                       complex_data=vh['complex_data']) as fw:
            new_header = fw.header0
            fw.write(dv_data)

        assert self.get_vdif_header(new_header) == vh
        with dada.open(dada_file, 'rs') as fh:
            header = fh.header0
            new_dada_data = fh.read()
        assert header == new_header
        assert self.get_vdif_header(header) == vh
        assert np.allclose(new_dada_data, dada_data)



# Licensed under the GPLv3 - see LICENSE
"""Green Bank Ultimate Pulsar Processing Instrument (GUPPI) format
reader/writer.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from .base import open
from .header import GUPPIHeader
from .payload import GUPPIPayload
from .frame import GUPPIFrame



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import importlib
import pytest
from astropy.time import Time

from .. import file_info
from ..data import (SAMPLE_MARK4 as SAMPLE_M4, SAMPLE_MARK5B as SAMPLE_M5B,
                    SAMPLE_VDIF, SAMPLE_MWA_VDIF as SAMPLE_MWA, SAMPLE_DADA,
                    SAMPLE_PUPPI, SAMPLE_GSB_RAWDUMP_HEADER,
                    SAMPLE_GSB_RAWDUMP, SAMPLE_GSB_PHASED_HEADER,
                    SAMPLE_GSB_PHASED)


@pytest.mark.parametrize(
    ('sample', 'format_'),
    ((SAMPLE_M4, 'mark4'),
     (SAMPLE_M5B, 'mark5b'),
     (SAMPLE_VDIF, 'vdif'),
     (SAMPLE_MWA, 'vdif'),
     (SAMPLE_DADA, 'dada'),
     (SAMPLE_PUPPI, 'guppi'),
     (SAMPLE_GSB_RAWDUMP_HEADER, 'gsb'),
     (SAMPLE_GSB_PHASED_HEADER, 'gsb')))
def test_basic_file_info(sample, format_):
    info = file_info(sample)
    info_dict = info()
    assert info.format == format_
    assert info_dict['format'] == format_
    if format_.startswith('mark') or format_.startswith('gsb'):
        assert info.missing
        assert 'missing' in info_dict
    else:
        assert not info.missing
        assert 'missing' not in info_dict


@pytest.mark.parametrize(
    ('sample', 'missing'),
    ((SAMPLE_M4, {'decade', 'ref_time'}),
     (SAMPLE_M5B, {'kday', 'ref_time', 'nchan'})))
def test_open_missing_args(sample, missing):
    info = file_info(sample)
    assert info.missing
    assert set(info.missing) == missing


@pytest.mark.parametrize(
    ('sample', 'format_', 'used', 'consistent', 'inconsistent'),
    ((SAMPLE_M4, 'mark4', ('ref_time',), ('nchan',), ()),
     (SAMPLE_M5B, 'mark5b', ('ref_time', 'nchan'), (), ()),
     (SAMPLE_VDIF, 'vdif', (), ('nchan', 'ref_time'), ()),
     (SAMPLE_DADA, 'dada', (), ('ref_time',), ('nchan',)),
     (SAMPLE_PUPPI, 'guppi', (), ('nchan',), ('ref_time',))))
def test_file_info(sample, format_, used, consistent, inconsistent):
    # Pass on extra arguments needed to get Mark4 and Mark5B to pass.
    # For GSB, we also need raw files, so we omit them.
    extra_args = {'ref_time': Time('2014-01-01'),
                  'nchan': 8}
    info = file_info(sample, **extra_args)
    assert info.format == format_
    assert not info.missing
    info_dict = info()
    for attr in info.attr_names:
        assert getattr(info, attr) is not None
        assert attr in info_dict
    assert set(info.used_kwargs.keys()) == set(used)
    assert set(info.consistent_kwargs.keys()) == set(consistent)
    assert set(info.inconsistent_kwargs.keys()) == set(inconsistent)
    assert set(info.irrelevant_kwargs.keys()) == set()
    # Check that extraneous arguments get classified correctly.
    info2 = file_info(sample, life=42, **extra_args)
    assert info2.used_kwargs == info.used_kwargs
    assert info2.consistent_kwargs == info.consistent_kwargs
    assert info2.inconsistent_kwargs == info.inconsistent_kwargs
    assert info2.irrelevant_kwargs == {'life': 42}
    # Check we can indeed open a file with the extra arguments.
    module = importlib.import_module('.' + info.format, package='baseband')
    with module.open(sample, mode='rs', **info.used_kwargs) as fh:
        info3 = fh.info
    assert info3() == info_dict


@pytest.mark.parametrize(
    ('sample', 'raw', 'mode'),
    ((SAMPLE_GSB_RAWDUMP_HEADER, SAMPLE_GSB_RAWDUMP, 'rawdump'),
     (SAMPLE_GSB_PHASED_HEADER, SAMPLE_GSB_PHASED, 'phased')))
def test_gsb_with_raw_files(sample, raw, mode):
    info = file_info(sample, raw=raw)
    assert info.format == 'gsb'
    assert not info.missing
    module = importlib.import_module('.' + info.format, package='baseband')
    # Check we can indeed open a file with the extra arguments.
    with module.open(sample, mode='rs', **info.used_kwargs) as fh:
        info2 = fh.info
    assert info2() == info()



# Licensed under the GPLv3 - see LICENSE
"""Mark5B VLBI data reader.

Code inspired by Walter Brisken's mark5access.  See
https://github.com/demorest/mark5access.

Also, for the Mark5B design, see
http://www.haystack.mit.edu/tech/vlbi/mark5/mark5_memos/019.pdf
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from .base import open
from .header import Mark5BHeader
from .payload import Mark5BPayload
from .frame import Mark5BFrame



# Licensed under the GPLv3 - see LICENSE
"""
Base definitions for VLBI Headers, used for VDIF and Mark 5B.

Defines a header class VLBIHeaderBase that can be used to hold the words
corresponding to a frame header, providing access to the values encoded in
via a dict-like interface.  Definitions for headers are constructed using
the HeaderParser class.
"""
from __future__ import absolute_import, division, print_function
from copy import copy
import struct
import warnings
from collections import OrderedDict
import numpy as np


__all__ = ['four_word_struct', 'eight_word_struct',
           'make_parser', 'make_setter',
           'HeaderProperty', 'HeaderPropertyGetter',
           'HeaderParser', 'VLBIHeaderBase']

four_word_struct = struct.Struct('<4I')
"""Struct instance that packs/unpacks 4 unsigned 32-bit integers."""
eight_word_struct = struct.Struct('<8I')
"""Struct instance that packs/unpacks 8 unsigned 32-bit integers."""


def make_parser(word_index, bit_index, bit_length, default=None):
    """Construct a function that converts specific bits from a header.

    The function acts on a tuple/array of 32-bit words, extracting given bits
    from a specific word and convert them to bool (for single bit) or integer.

    The parameters are those that define header keywords, and all parsers do
    ``(words[word_index] >> bit_index) & ((1 << bit_length) - 1)``, except that
    that they have been optimized for the specific cases of single bits,
    full words, and items starting at bit 0.  As a special case, bit_length=64
    allows one to extract two words as a single (long) integer.

    Parameters
    ----------
    word_index : int
        Index into the tuple of words passed to the function.
    bit_index : int
        Index to the starting bit of the part to be extracted.
    bit_length : int
        Number of bits to be extracted.

    Returns
    -------
    parser : function
        To be used as ``parser(words)``.
    """
    if bit_length == 1:
        def parser(words):
            return (words[word_index] & (1 << bit_index)) != 0

    elif bit_length == 32:
        assert bit_index == 0

        def parser(words):
            return words[word_index]

    elif bit_length == 64:
        assert bit_index == 0

        def parser(words):
            return words[word_index] + words[word_index + 1] * (1 << 32)

    else:
        bit_mask = (1 << bit_length) - 1  # e.g., bit_length=8 -> 0xff
        if bit_index == 0:
            def parser(words):
                return words[word_index] & bit_mask

        else:
            def parser(words):
                return (words[word_index] >> bit_index) & bit_mask

    return parser


def make_setter(word_index, bit_index, bit_length, default=None):
    """Construct a function that uses a value to set specific bits in a header.

    The function will act on a tuple/array of words, setting given bits
    from a given word using a value.

    The parameters are just those that define header keywords.

    Parameters
    ----------
    word_index : int
        Index into the tuple of words passed to the function.
    bit_index : int
        Index to the starting bit of the part to be extracted.
    bit_length : int
        Number of bits to be extracted.
    default : int or bool or None
        Possible default value to use in function if no default is passed on.

    Returns
    -------
    setter : function
        To be used as ``setter(words, value)``.
    """
    def setter(words, value):
        if value is None and default is not None:
            value = default
        bit_mask = (1 << bit_length) - 1
        # Check that value will fit within the bit limits.
        if np.any(value & bit_mask != value):
            raise ValueError("{0} cannot be represented with {1} bits"
                             .format(value, bit_length))
        if bit_length == 64:
            word1 = value & (1 << 32) - 1
            word2 = value >> 32
            words[word_index] = word1
            words[word_index + 1] = word2
            return words

        word = words[word_index]
        # Zero the part to be set, and add the value.
        bit_mask <<= bit_index
        word = ((word | bit_mask) ^ bit_mask) | (value << bit_index)
        words[word_index] = word
        return words

    return setter


def get_default(word_index, bit_index, bit_length, default=None):
    return default


class HeaderProperty(object):
    """Mimic a dictionary, calculating entries from header words.

    Used to calculate setter functions and extract default values.

    Parameters
    ----------
    header_parser : `HeaderParser`
        A dict with header encoding information.
    getter : function
        Function that uses the encoding information to calculate a result.
    """
    def __init__(self, header_parser, getter, doc=None):
        self.header_parser = header_parser
        self.getter = getter
        if doc is not None:
            self.__doc__ = doc

    def __getitem__(self, item):
        definition = self.header_parser[item]
        return self.getter(*definition)


class HeaderPropertyGetter(object):
    """Special property for attaching HeaderProperty."""
    def __init__(self, getter, doc=None):
        self.getter = getter
        self.__doc__ = doc or getter.__doc__

    def __get__(self, instance, owner_cls=None):
        if instance is None:  # pragma: no cover
            return self
        return HeaderProperty(instance, getattr(instance, self.getter),
                              doc=self.__doc__)


class HeaderParser(OrderedDict):
    """Parser & setter for VLBI header keywords.

    An ordered dict of header keywords, with values that describe how they are
    encoded in a given VLBI header.  Initialisation is as a normal OrderedDict,
    with a key, value pairs.  The value should be a tuple containing:

    word_index : int
        Index into the header words for this key.
    bit_index : int
        Index to the starting bit of the part used for this key.
    bit_length : int
        Number of bits.
    default : int or bool or None
        Possible default value to use in initialisation (e.g., a sync pattern).

    The class provides dict-like properties ``parsers``, ``setters``, and
    ``defaults``, which return functions that get a given keyword from header
    words, set the corresponding part of the header words to a value, or
    return the default value (if defined).

    Note that while in principle, parsers and setters could be calculated on
    the fly, we precalculate the parsers to speed up header keyword access.
    """

    def __init__(self, *args, **kwargs):
        self._make_parser = kwargs.pop('make_parser', make_parser)
        self._make_setter = kwargs.pop('make_setter', make_setter)
        self._get_default = kwargs.pop('get_default', get_default)
        # Use a dict rather than OrderedDict for the parsers for better speed.
        # Note that this gets filled by calls to __setitem__.
        self._parsers = {}
        super(HeaderParser, self).__init__(*args, **kwargs)

    def copy(self):
        """Make an independent copy."""
        return self.__class__(self, make_parser=self._make_parser,
                              make_setter=self._make_setter,
                              get_default=self._get_default)

    def __add__(self, other):
        if not isinstance(other, HeaderParser):
            return NotImplemented
        result = self.copy()
        result.update(other)
        return result

    def __setitem__(self, item, value):
        self._parsers[item] = self._make_parser(*value)
        super(HeaderParser, self).__setitem__(item, value)

    @property
    def parsers(self):
        """Dict with functions to get specific header values."""
        return self._parsers

    defaults = HeaderPropertyGetter(
        '_get_default',
        doc="Dict-like allowing access to default header values.")

    setters = HeaderPropertyGetter(
        '_make_setter',
        doc="Dict-like returning function to set specific header value.")

    def update(self, other):
        """Update the parser with the information from another one."""
        if not isinstance(other, HeaderParser):
            raise TypeError("can only update using a HeaderParser instance.")
        super(HeaderParser, self).update(other)
        # Update the parsers rather than recalculate all the functions.
        self._parsers.update(other._parsers)


class VLBIHeaderBase(object):
    """Base class for all VLBI headers.

    Defines a number of common routines.

    Generally, the actual class should define:

      _struct: `~struct.Struct` instance that can pack/unpack header words.

      _header_parser: `HeaderParser` instance corresponding to this class.

      _properties: tuple of properties accessible/usable in initialisation

    It also should define properties (getters *and* setters):

      payload_nbytes: number of bytes used by payload

      frame_nbytes: total number of bytes for header + payload

      get_time, set_time, and a corresponding time property:
           time at start of payload

    Parameters
    ----------
    words : tuple or list of int, or None
        header words (generally, 32 bit unsigned int).  If `None`,
        set to a list of zeros for later initialisation.  If given as a tuple,
        the header is immutable.
    verify : bool
        Whether to do basic verification of integrity.  For the base class,
        checks that the number of words is consistent with the struct size.
    """

    _properties = ('payload_nbytes', 'frame_nbytes', 'time')
    """Properties accessible/usable in initialisation for all headers."""

    def __init__(self, words, verify=True):
        if words is None:
            self.words = [0] * (self._struct.size // 4)
        else:
            self.words = words
        if verify:
            self.verify()

    def verify(self):
        """Verify that the length of the words is consistent.

        Subclasses should override this to do more thorough checks.
        """
        assert len(self.words) == (self._struct.size // 4)

    def copy(self, **kwargs):
        """Create a mutable and independent copy of the header.

        Keyword arguments can be passed on as needed by possible subclasses.
        """
        kwargs.setdefault('verify', False)
        new = self.__class__(copy(self.words), **kwargs)
        new.mutable = True
        return new

    def __copy__(self):
        return self.copy()

    @property
    def nbytes(self):
        """Size of the header in bytes."""
        return self._struct.size

    @property
    def mutable(self):
        """Whether the header can be modified."""
        if isinstance(self.words, tuple):
            return False
        word0 = self.words[0]
        try:
            self.words[0] = 0
        except Exception:
            return False
        else:
            self.words[0] = word0
            return True

    @mutable.setter
    def mutable(self, mutable):
        if isinstance(self.words, np.ndarray):
            self.words.flags['WRITEABLE'] = mutable
        elif isinstance(self.words, tuple):
            if mutable:
                self.words = list(self.words)
        elif isinstance(self.words, list):
            if not mutable:
                self.words = tuple(self.words)
        else:
            raise TypeError("do not know how to set mutability of '.words' "
                            "of class {0}".format(type(self.words)))

    @classmethod
    def fromfile(cls, fh, *args, **kwargs):
        """Read VLBI Header from file.

        Arguments are the same as for class initialisation.  The header
        constructed will be immutable.
        """
        s = fh.read(cls._struct.size)
        if len(s) != cls._struct.size:
            raise EOFError
        return cls(cls._struct.unpack(s), *args, **kwargs)

    def tofile(self, fh):
        """Write VLBI frame header to filehandle."""
        return fh.write(self._struct.pack(*self.words))

    @classmethod
    def fromvalues(cls, *args, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header = cls(<words>)``, ``cls.fromvalues(**header) == header``.

        However, unlike for the `fromkeys` class method, data can also be set
        using arguments named after header methods, such as ``time``.

        Parameters
        ----------
        *args
            Possible arguments required to initialize an empty header.
        **kwargs
            Values used to initialize header keys or methods.
        """
        # Initialize an empty header, and update it with the keyword arguments.
        self = cls(None, *args, verify=False)
        # Set defaults in keyword arguments.
        for key in set(self.keys()).difference(kwargs.keys()):
            default = self._header_parser.defaults[key]
            if default is not None:
                kwargs[key] = default

        self.update(**kwargs)
        return self

    @classmethod
    def fromkeys(cls, *args, **kwargs):
        """Initialise a header from parsed values.

        Like fromvalues, but without any interpretation of keywords.

        Raises
        ------
        KeyError : if not all keys required are present in ``kwargs``
        """
        self = cls(None, *args, verify=False)
        not_in_both = (set(self.keys()).symmetric_difference(kwargs) -
                       {'verify'})
        if not_in_both:
            not_in_kwarg = set(self.keys()).difference(kwargs)
            not_in_self = set(kwargs).difference(self.keys()) - {'verify'}
            msg_parts = []
            for item, msg in ((not_in_kwarg, "is missing keywords ({0})"),
                              (not_in_self, "contains extra keywords ({0})")):
                if item:
                    msg_parts.append(msg.format(item))

            raise KeyError("input list " + " and ".join(msg_parts))

        self.update(**kwargs)
        return self

    def update(self, **kwargs):
        """Update the header by setting keywords or properties.

        Here, any keywords matching header keys are applied first, and any
        remaining ones are used to set header properties, in the order set
        by the class (in ``_properties``).

        Parameters
        ----------
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        verify = kwargs.pop('verify', True)

        # First use keywords which are also keys into self.
        for key in set(kwargs.keys()).intersection(self.keys()):
            self[key] = kwargs.pop(key)

        # Next, use remaining keyword arguments to set properties.
        # Order is important, so we cannot use an intersection as above.
        if kwargs:
            for key in self._properties:
                if key in kwargs:
                    setattr(self, key, kwargs.pop(key))

            if kwargs:
                warnings.warn("some keywords unused in header update: {0}"
                              .format(kwargs))

        if verify:
            self.verify()

    def __getitem__(self, item):
        """Get the value a particular header item from the header words."""
        try:
            return self._header_parser._parsers[item](self.words)
        except KeyError:
            raise KeyError("{0} header does not contain {1}"
                           .format(self.__class__.__name__, item))

    def __setitem__(self, item, value):
        """Set the value of a particular header item in the header words."""
        try:
            self._header_parser.setters[item](self.words, value)
        except KeyError:
            raise KeyError("{0} header does not contain {1}"
                           .format(self.__class__.__name__, item))
        except(TypeError, ValueError):
            if not self.mutable:
                raise TypeError("header is immutable. Set '.mutable` attribute"
                                " or make a copy.")
            else:
                raise

    def keys(self):
        return self._header_parser.keys()

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.keys()

    def __contains__(self, key):
        return key in self.keys()

    def __eq__(self, other):
        return (type(self) is type(other) and
                np.all(np.array(self.words, copy=False) ==
                       np.array(other.words, copy=False)))

    @staticmethod
    def _repr_as_hex(key):
        return (key.startswith('bcd') or key.startswith('crc') or
                key == 'sync_pattern')

    def __repr__(self):
        name = self.__class__.__name__
        return ("<{0} {1}>".format(name, (",\n  " + len(name) * " ").join(
            ["{0}: {1}".format(k, hex(self[k]) if self._repr_as_hex(k)
                               else self[k])
             for k in self.keys()])))



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import pytest
from astropy.time import Time
import astropy.units as u

from .. import open as baseband_open, file_info
from ..data import (SAMPLE_MARK4 as SAMPLE_M4, SAMPLE_MARK5B as SAMPLE_M5B,
                    SAMPLE_VDIF, SAMPLE_MWA_VDIF as SAMPLE_MWA, SAMPLE_DADA)


@pytest.mark.parametrize(
    ('sample', 'fmt'),
    ((SAMPLE_M4, 'mark4'),
     (SAMPLE_M5B, 'mark5b'),
     (SAMPLE_VDIF, 'vdif')))
def test_open(sample, fmt):
    # Use extra arguments that allow all formats to succeed.
    extra_args = {'nchan': 8,
                  'ref_time': Time('2014-01-01'),
                  'sample_rate': 32.*u.MHz}
    info = file_info(sample, fmt, **extra_args)
    with baseband_open(sample, 'rs', **extra_args) as fh:
        assert fh.start_time == info.start_time


@pytest.mark.parametrize('sample', (SAMPLE_M4, SAMPLE_M5B))
def test_open_missing_args(sample):
    with pytest.raises(TypeError) as exc:
        baseband_open(sample, 'rs')
    assert "missing required arguments" in str(exc.value)


def test_open_wrong_args():
    # Use extra arguments that allow formats to succeed up to the stream
    # reader, but then fail on an incorrect sample rate.
    mark4_args = {'nchan': 8,
                  'ref_time': Time('2014-01-01')}
    with pytest.raises(ValueError) as exc:  # wrong sample_rate
        baseband_open(SAMPLE_M4, 'rs', sample_rate=31*u.MHz, **mark4_args)
    assert "inconsistent" in str(exc.value)

    with pytest.raises(TypeError) as exc:  # extraneous argument
        baseband_open(SAMPLE_M4, 'rs', life=42, **mark4_args)
    assert "unexpected" in str(exc.value)

    with pytest.raises(ValueError) as exc:  # wrong decade
        baseband_open(SAMPLE_VDIF, 'rs', decade=2000)
    assert "inconsistent" in str(exc.value)

    with pytest.raises(ValueError) as exc:  # wrong kday
        baseband_open(SAMPLE_VDIF, 'rs', kday=55000)
    assert "inconsistent" in str(exc.value)

    with pytest.raises(ValueError) as exc:  # ref_time too far off.
        baseband_open(SAMPLE_VDIF, 'rs', ref_time=Time('J2000'))
    assert "inconsistent" in str(exc.value)

    with pytest.raises(ValueError) as exc:  # inconsistent nchan.
        baseband_open(SAMPLE_DADA, 'rs', nchan=8)
    assert "inconsistent" in str(exc.value)


def test_open_write_checks():
    # Cannot have multiple formats for writing.
    with pytest.raises(ValueError):
        baseband_open('a.a', 'wb', fmt=('dada', 'mark4'))



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..vlbi_base.file_info import VLBIFileReaderInfo


class Mark4FileReaderInfo(VLBIFileReaderInfo):
    """Standardized information on Mark 4 file readers.

    The ``info`` descriptor has a number of standard attributes, which are
    determined from arguments passed in opening the file, from the first header
    (``info.header0``) and from possibly scanning the file to determine the
    duration of frames.  `Mark4FileReaderInfo` has two additional attributes
    specific to Mark 4 files (``ntrack`` and ``offset0``, see below).

    Attributes
    ----------
    format : str or `None`
        File format, or `None` if the underlying file cannot be parsed.
    frame_rate : `~astropy.units.Quantity`
        Number of data frames per unit of time.
    sample_rate : `~astropy.units.Quantity`
        Complete samples per unit of time.
    samples_per_frame : int
        Number of complete samples in each frame.
    sample_shape : tuple
        Dimensions of each complete sample (e.g., ``(nchan,)``).
    bps : int
        Number of bits used to encode each elementary sample.
    complex_data : bool
        Whether the data are complex.
    start_time : `~astropy.time.Time`
        Time of the first complete sample.
    ntrack : int
        Number of "tape tracks" simulated in the disk file.
    offset0 : int
        Offset in bytes from the start of the file to the location of the
        first header.
    missing : dict
        Entries are keyed by names of arguments that should be passed to
        the file reader to obtain full information. The associated entries
        explain why these arguments are needed. For Mark 4, the possible
        entries are ``decade`` and ``ref_time``.

    Examples
    --------
    The most common use is simply to print information::

        >>> from baseband.data import SAMPLE_MARK4
        >>> from baseband import mark4
        >>> fh = mark4.open(SAMPLE_MARK4, 'rb')
        >>> fh.info
        File information:
        format = mark4
        frame_rate = 400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 80000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        offset0 = 2696
        <BLANKLINE>
        missing:  decade, ref_time: needed to infer full times.
        <BLANKLINE>
        >>> fh.close()

        >>> fh = mark4.open(SAMPLE_MARK4, 'rb', decade=2010)
        >>> fh.info
        File information:
        format = mark4
        frame_rate = 400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 80000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        start_time = 2014-06-16T07:38:12.475000000
        offset0 = 2696
        >>> fh.close()
    """
    attr_names = VLBIFileReaderInfo.attr_names + ('ntrack', 'offset0')
    _header0_attrs = ('bps', 'samples_per_frame')
    _parent_attrs = ('ntrack', 'decade', 'ref_time')

    def _get_header0(self):
        fh = self._parent
        old_offset = fh.tell()
        try:
            fh.seek(0)
            offset0 = fh.locate_frame()
            if offset0 is None:
                return None

            self.offset0 = offset0
            return fh.read_header()
        except Exception:
            return None
        finally:
            fh.seek(old_offset)

    def _collect_info(self):
        super(Mark4FileReaderInfo, self)._collect_info()
        if self:
            self.complex_data = False
            # TODO: Shouldn't Mark4Header provide this?
            self.sample_shape = (self.header0.nchan,)
            if self.decade is None and self.ref_time is None:
                self.missing['decade'] = self.missing['ref_time'] = (
                    "needed to infer full times.")



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..vlbi_base.frame import VLBIFrameBase
from .header import GUPPIHeader
from .payload import GUPPIPayload


__all__ = ['GUPPIFrame']


class GUPPIFrame(VLBIFrameBase):
    """Representation of a GUPPI file, consisting of a header and payload.

    Parameters
    ----------
    header : `~baseband.guppi.GUPPIHeader`
        Wrapper around the header lines, providing access to the values.
    payload : `~baseband.guppi.GUPPIPayload`
        Wrapper around the payload, provding mechanisms to decode it.
    valid : bool, optional
        Whether the data are valid.  Default: `True`.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Notes
    -----
    GUPPI files do not support storing whether data are valid or not on disk.
    Hence, this has to be determined independently.  If ``valid=False``, any
    decoded data are set to ``cls.fill_value`` (by default, 0).

    The Frame can also be instantiated using class methods:

      fromfile : read header and and map or read payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    One can decode part of the payload by indexing or slicing the frame.

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """
    _header_class = GUPPIHeader
    _payload_class = GUPPIPayload

    @classmethod
    def fromfile(cls, fh, memmap=True, valid=True, verify=True):
        """Read a frame from a filehandle, possible mapping the payload.

        Parameters
        ----------
        fh : filehandle
            To read header from.
        memmap : bool, optional
            If `True` (default), use `~numpy.memmap` to map the payload.
            If `False`, just read it from disk.
        valid : bool, optional
            Whether the data are valid (default: `True`). Note that this cannot
            be inferred from the header or payload itself.  If `False`, any
            data read will be set to ``cls.fill_value``.
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.
        """
        header = cls._header_class.fromfile(fh, verify=verify)
        payload = cls._payload_class.fromfile(fh, header=header, memmap=memmap)
        return cls(header, payload, valid=valid, verify=verify)

    @classmethod
    def fromdata(cls, data, header=None, valid=True, verify=True, **kwargs):
        """Construct frame from data and header.

        Note that since GUPPI files are generally very large, one would
        normally map the file, and then set pieces of it by assigning to slices
        of the frame.  See `~baseband.guppi.base.GUPPIFileWriter.memmap_frame`.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding complex or real data to be encoded.
        header : `~baseband.guppi.GUPPIHeader` or None, optional
            If not given, will attempt to generate one using the keywords.
        valid : bool, optional
            Whether the data are valid (default: `True`). Note that this
            information cannot be written to disk.
        verify : bool, optional
            Whether or not to do basic assertions that check the integrity.
            Default: `True`.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if header is None:
            header = cls._header_class.fromvalues(verify=verify, **kwargs)
        payload = cls._payload_class.fromdata(data, header=header)
        return cls(header, payload, valid=valid, verify=verify)



# Licensed under the GPLv3 - see LICENSE
from __future__ import division, unicode_literals, print_function

import astropy.units as u
from astropy.utils import lazyproperty

from ..helpers import sequentialfile as sf
from ..vlbi_base.base import (make_opener, VLBIFileBase, VLBIFileReaderBase,
                              VLBIStreamBase, VLBIStreamReaderBase,
                              VLBIStreamWriterBase)
from .header import GUPPIHeader
from .payload import GUPPIPayload
from .frame import GUPPIFrame


__all__ = ['GUPPIFileReader', 'GUPPIFileWriter', 'GUPPIStreamBase',
           'GUPPIStreamReader', 'GUPPIStreamWriter', 'open']


class GUPPIFileReader(VLBIFileReaderBase):
    """Simple reader for GUPPI files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `read_frame` and `get_frame_rate`. By default, frame payloads
    are mapped rather than fully read into physical memory.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.
    """

    def read_header(self):
        """Read a single header from the file.

        Returns
        -------
        header : `~baseband.guppi.GUPPIHeader`
        """
        return GUPPIHeader.fromfile(self.fh_raw)

    def read_frame(self, memmap=True, verify=True):
        """Read the frame header and read or map the corresponding payload.

        Parameters
        ----------
        memmap : bool, optional
            If `True` (default), map the payload using `~numpy.memmap`, so that
            parts are only loaded into memory as needed to access data.
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.

        Returns
        -------
        frame : `~baseband.guppi.GUPPIFrame`
            With ``.header`` and ``.payload`` properties.  The ``.data``
            property returns all data encoded in the frame.  Since this may
            be too large to fit in memory, it may be better to access the
            parts of interest by slicing the frame.
        """
        return GUPPIFrame.fromfile(self.fh_raw, memmap=memmap, verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        The routine uses the sample rate and number of samples per frame
        (excluding overlap) from the first header in the file.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        oldpos = self.tell()
        self.seek(0)
        try:
            header = self.read_header()
            return (header.sample_rate /
                    (header.samples_per_frame - header.overlap)).to(u.Hz)
        finally:
            self.seek(oldpos)


class GUPPIFileWriter(VLBIFileBase):
    """Simple writer/mapper for GUPPI files.

    Adds `write_frame` and `memmap_frame` methods to the VLBI binary file
    wrapper.  The latter allows one to encode data in pieces, writing to disk
    as needed.
    """
    def write_frame(self, data, header=None, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.guppi.GUPPIFrame`
            If an array, a ``header`` should be given, which will be used to
            get the information needed to encode the array, and to construct
            the GUPPI frame.
        header : `~baseband.guppi.GUPPIHeader`
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a `~baseband.guppi.GUPPIFrame` instance.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, GUPPIFrame):
            data = GUPPIFrame.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)

    def memmap_frame(self, header=None, **kwargs):
        """Get frame by writing the header to disk and mapping its payload.

        The header is written to disk immediately, but the payload is mapped,
        so that it can be filled in pieces, by setting slices of the frame.

        Parameters
        ----------
        header : `~baseband.guppi.GUPPIHeader`
            Written to disk immediately.  Can instead give keyword arguments to
            construct a header.
        **kwargs
            If ``header`` is not given, these are used to initialize one.

        Returns
        -------
        frame: `~baseband.guppi.GUPPIFrame`
            By assigning slices to data, the payload can be encoded piecewise.
        """
        if header is None:
            header = GUPPIHeader.fromvalues(**kwargs)
        header.tofile(self.fh_raw)
        payload = GUPPIPayload.fromfile(self.fh_raw, memmap=True,
                                        header=header)
        return GUPPIFrame(header, payload)


class GUPPIStreamBase(VLBIStreamBase):
    """Base for GUPPI streams."""

    _sample_shape_maker = GUPPIPayload._sample_shape_maker

    def __init__(self, fh_raw, header0, squeeze=True, subset=(), verify=True):

        # GUPPI headers report their offsets using 'PKTIDX', the number of
        # unique UDP data packets (i.e. excluding overlap) written since the
        # start of observation.  'PKTSIZE' is the packet size in bytes.  Here
        # we calculate the packets per frame.
        self._packets_per_frame = (
            (header0.payload_nbytes - header0.overlap * header0._bpcs // 8) //
            header0['PKTSIZE'])

        # Set samples per frame to unique ones, excluding overlap.
        samples_per_frame = header0.samples_per_frame - header0.overlap

        super(GUPPIStreamBase, self).__init__(
            fh_raw=fh_raw, header0=header0, sample_rate=header0.sample_rate,
            samples_per_frame=samples_per_frame,
            unsliced_shape=header0.sample_shape, bps=header0.bps,
            complex_data=header0.complex_data, squeeze=squeeze, subset=subset,
            fill_value=0., verify=verify)

    # Overriding so the docstring indicates the exclusion of the overlap.
    samples_per_frame = property(VLBIStreamBase.samples_per_frame.fget,
                                 VLBIStreamBase.samples_per_frame.fset,
                                 doc=("Number of complete samples per frame, "
                                      "excluding overlap."))


class GUPPIStreamReader(GUPPIStreamBase, VLBIStreamReaderBase):
    """GUPPI format reader.

    Allows access to GUPPI files as a continuous series of samples.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw GUPPI stream.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object or tuple of objects, optional
        Specific components of the complete sample to decode (after possibly
        squeezing).  If a single indexing object is passed, it selects
        polarizations.  With a tuple, the first selects polarizations and the
        second selects channels.  If the tuple is empty (default), all
        components are read.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frame of the stream is always checked, so ``verify`` is effective only
        when reading sequences of files.  Default: `True`.
    """
    def __init__(self, fh_raw, squeeze=True, subset=(), verify=True):
        fh_raw = GUPPIFileReader(fh_raw)
        header0 = GUPPIHeader.fromfile(fh_raw)
        super(GUPPIStreamReader, self).__init__(fh_raw, header0,
                                                squeeze=squeeze,
                                                subset=subset, verify=verify)

    @lazyproperty
    def _last_header(self):
        """Header of the last file for this stream."""
        # Seek forward rather than backward, as last frame often has missing
        # bytes.
        nframes, fframe = divmod(self.fh_raw.seek(0, 2),
                                 self.header0.frame_nbytes)
        self.fh_raw.seek((nframes - 1) * self.header0.frame_nbytes)
        return self.fh_raw.read_header()

    def _read_frame(self, index):
        self.fh_raw.seek(index * self.header0.frame_nbytes)
        frame = self.fh_raw.read_frame(memmap=True, verify=self.verify)
        assert (frame.header['PKTIDX'] - self.header0['PKTIDX'] ==
                index * self._packets_per_frame)
        return frame


class GUPPIStreamWriter(GUPPIStreamBase, VLBIStreamWriterBase):
    """GUPPI format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    raw : filehandle
        For writing the header and raw data to storage.
    header0 : :class:`~baseband.guppi.GUPPIHeader`
        Header for the first frame, holding time information, etc.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input,
        and adds any dimensions of length unity.
    """
    def __init__(self, fh_raw, header0, squeeze=True):
        assert header0.get('OVERLAP', 0) == 0, ("overlap must be 0 when "
                                                "writing GUPPI files.")
        fh_raw = GUPPIFileWriter(fh_raw)
        super(GUPPIStreamWriter, self).__init__(fh_raw, header0,
                                                squeeze=squeeze)

    def _make_frame(self, index):
        header = self.header0.copy()
        header.update(pktidx=self.header0['PKTIDX'] +
                      index * self._packets_per_frame)
        return self.fh_raw.memmap_frame(header)

    def _write_frame(self, frame, valid=True):
        assert frame is self._frame
        frame.valid = valid
        # Deleting frame flushes memmap'd data to disk.
        # (Of course, this gets deleted automatically when going out of
        # scope, and furthermore the link in self._frame will still exist
        # -- it only gets deleted in VLBIStreamWriter.write)
        del frame


opener = make_opener('GUPPI', globals(), doc="""
--- For reading a stream : (see `~baseband.guppi.base.GUPPIStreamReader`)

squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object or tuple of objects, optional
    Specific components of the complete sample to decode (after possibly
    squeezing).  If a single indexing object is passed, it selects
    polarizations.  With a tuple, the first selects polarizations and the
    second selects channels.  If the tuple is empty (default), all
    components are read.

--- For writing a stream : (see `~baseband.guppi.base.GUPPIStreamWriter`)

header0 : `~baseband.guppi.GUPPIHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input, and adds
    any dimensions of length unity.
frames_per_file : int, optional
    When writing to a sequence of files, sets the number of frames
    within each file.  Default: 128.
**kwargs
    If the header is not given, an attempt will be made to construct one
    with any further keyword arguments.

--- Header keywords : (see :meth:`~baseband.guppi.GUPPIHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.  Must have an integer number of seconds.
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each
    channel of each polarization is sampled.
samples_per_frame : int
    Number of complete samples per frame.  Can alternatively give
    ``payload_nbytes``.
payload_nbytes : int
    Number of bytes per payload.  Can alternatively give ``samples_per_frame``.
offset : `~astropy.units.Quantity` or `~astropy.time.TimeDelta`, optional
    Time offset from the start of the whole observation (default: 0).
npol : int, optional
    Number of polarizations (default: 1).
nchan : int, optional
    Number of channels (default: 1).  For GUPPI, complex data is only allowed
    when nchan > 1.
bps : int, optional
    Bits per elementary sample, i.e. per real or imaginary component for
    complex data (default: 8).

Returns
-------
Filehandle
    :class:`~baseband.guppi.base.GUPPIFileReader` or
    :class:`~baseband.guppi.base.GUPPIFileWriter` (binary), or
    :class:`~baseband.guppi.base.GUPPIStreamReader` or
    :class:`~baseband.guppi.base.GUPPIStreamWriter` (stream).
""")


# Need to wrap the opener to be able to deal with file lists or templates.
# TODO: move this up to the opener??
def open(name, mode='rs', **kwargs):
    frames_per_file = kwargs.pop('frames_per_file', 128)
    header0 = kwargs.get('header0', None)
    squeeze = kwargs.pop('squeeze', None)
    # If sequentialfile object, check that it's opened properly.
    if isinstance(name, sf.SequentialFileBase):
        assert (('r' in mode and name.mode == 'rb') or
                ('w' in mode and name.mode == 'w+b')), (
                    "open only accepts sequential files opened in 'rb' mode "
                    "for reading or 'w+b' mode for writing.")
    is_sequence = isinstance(name, (tuple, list))

    if 'b' not in mode:
        if header0 is None:
            if 'w' in mode:
                # For writing a header is required.
                header0 = GUPPIHeader.fromvalues(**kwargs)
                kwargs = {}
            else:
                header0 = {}

        if is_sequence:
            if 'r' in mode:
                name = sf.open(name, 'rb')
            else:
                name = sf.open(name, 'w+b', file_size=(
                    frames_per_file * header0.frame_nbytes))

        if header0 and 'w' in mode:
            kwargs['header0'] = header0

    if squeeze is not None:
        kwargs['squeeze'] = squeeze

    return opener(name, mode, **kwargs)


open.__doc__ = opener.__doc__ + """\n
Notes
-----
For streams, one can also pass in a list of files, or equivalently a
`~baseband.helpers.sequentialfile` object (opened in 'rb' mode for reading or
'w+b' for writing).
"""



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark 5B payloads.

Implements a Mark5BPayload class used to store payload words, and decode to
or encode from a data array.

For the specification, see
http://www.haystack.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np
from collections import namedtuple
from ..vlbi_base.payload import VLBIPayloadBase
from ..vlbi_base.encoding import encode_2bit_base, decoder_levels


__all__ = ['init_luts', 'decode_2bit', 'encode_2bit',
           'Mark5BPayload']


# Some duplication with mark4.py here: lut2bit = mark4.lut2bit1
# Though lut1bit = -mark4.lut1bit, so perhaps not worth combining.
def init_luts():
    """Set up the look-up tables for levels as a function of input byte.

    For 1-bit mode, one has just the sign bit:
      === =====
       s  value
      === =====
       0  -1
       1  +1
      === =====

    For 2-bit mode, there is a sign and a magnitude, which encode:
     === === ===== =====
      m   s  value s*2+m
     === === ===== =====
      0   0  -Hi    0
      0   1  +1     2
      1   0  -1     1
      1   1  +Hi    3
     === === ===== =====

    See Table 13 in
    https://science.nrao.edu/facilities/vlba/publications/memos/upgrade/sensimemo13.pdf
    and
    http://www.haystack.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
    Appendix A: sign always on even bit stream (0, 2, 4, ...), and magnitude
    on adjacent odd stream (1, 3, 5, ...).

    In the above table, the last column is the index in the linearly increasing
    table of levels (``decoder_levels[2]``).
    """
    b = np.arange(256)[:, np.newaxis]
    sl = np.arange(8)
    lut1bit = decoder_levels[1][((b >> sl) & 1)]
    # 2-bit mode: sign bit in lower position thatn magnitude bit
    # ms=00,01,10,11 = -Hi, 1, -1, Hi (lut
    s = np.arange(0, 8, 2)  # 0, 2, 4, 6
    m = s + 1               # 1, 3, 5, 7
    sl = (((b >> s) & 1) << 1) + ((b >> m) & 1)
    lut2bit = decoder_levels[2][sl]
    return lut1bit, lut2bit


lut1bit, lut2bit = init_luts()


# def decode_1bit(frame, nvlbichan):
#     return lut1bit[frame].reshape(-1, nvlbichan)


# Decoders keyed by bits_per_sample, complex_data:
def decode_2bit(words):
    b = words.view(np.uint8)
    return lut2bit.take(b, axis=0)


shift2bit = np.arange(0, 8, 2).astype(np.uint8)
reorder = np.array([0, 2, 1, 3], dtype=np.uint8)


def encode_2bit(values):
    bitvalues = encode_2bit_base(values.reshape(-1, 4))
    # swap 1 & 2
    reorder.take(bitvalues, out=bitvalues)
    bitvalues <<= shift2bit
    return np.bitwise_or.reduce(bitvalues, axis=-1)


encode_2bit.__doc__ = encode_2bit_base.__doc__


class Mark5BPayload(VLBIPayloadBase):
    """Container for decoding and encoding VDIF payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    nchan : int, optional
        Number of channels.   Default: 1.
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    """

    _nbytes = 2500 * 4
    _encoders = {2: encode_2bit}
    _decoders = {2: decode_2bit}

    _sample_shape_maker = namedtuple('SampleShape', 'nchan')

    def __init__(self, words, nchan=1, bps=2, complex_data=False):
        if complex_data:
            raise ValueError("Mark5B format does not support complex data.")

        super(Mark5BPayload, self).__init__(words, sample_shape=(nchan,),
                                            bps=bps, complex_data=False)

    @classmethod
    def fromdata(cls, data, bps=2):
        """Encode data as payload, using a given number of bits per sample.

        It is assumed that the last dimension is the number of channels.
        """
        if data.dtype.kind == 'c':
            raise ValueError("Mark5B format does not support complex data.")
        encoder = cls._encoders[bps]
        words = encoder(data).view(cls._dtype_word)
        return cls(words, nchan=data.shape[-1], bps=bps)



# Licensed under the GPLv3 - see LICENSE
"""Provide a base class for "info" properties.

Loosely based on `~astropy.utils.data_info.DataInfo`.
"""
from __future__ import division, unicode_literals, print_function

import astropy.units as u
from astropy.extern import six


class VLBIInfoMeta(type):
    # Ensure all attributes are initialized to None, so that they are
    # always available (do this rather than overwrite __getattr__ so that
    # we can generate docstrings in sphinx for them).
    def __init__(cls, name, bases, dct):
        super(VLBIInfoMeta, cls).__init__(name, bases, dct)
        attr_names = dct.get('attr_names', ())
        for attr in attr_names:
            setattr(cls, attr, None)


@six.add_metaclass(VLBIInfoMeta)
class VLBIInfoBase(object):
    """Container providing a standardized interface to file information."""

    attr_names = ('format',)
    """Attributes that the container provides."""

    _parent_attrs = ()
    _parent = None

    def _collect_info(self):
        # We link to attributes from the parent rather than just overriding
        # __getattr__ to allow us to look for changes.
        for attr in self._parent_attrs:
            setattr(self, attr, getattr(self._parent, attr))
        self.missing = {}

    def _up_to_date(self):
        """Determine whether the information we have stored is up to date."""
        return all(getattr(self, attr) == getattr(self._parent, attr)
                   for attr in self._parent_attrs)

    def __get__(self, instance, owner_cls):
        if instance is None:
            # Unbound descriptor, nothing to do.
            return self

        # Check if we have a stored and up to date copy.
        info = instance.__dict__.get('info')
        if info is None or not info._up_to_date():
            # If not, create a new instance and fill it.  Notes:
            # - We cannot change "self", as this was created on the class.
            # - We start from scratch rather than determine what is no longer
            #   up to date, since we cannot know what an update may influence
            #   (e.g., for Mark 4, a change in ref_time affect start_time).
            info = instance.__dict__['info'] = self.__class__()
            info._parent = instance
            info._collect_info()

        return info

    def __set__(self, info):
        # We do need to define __set__ since this ensures we are treated as
        # a "data descriptor", i.e., that our __get__ will get called even
        # if "info" is present in instance.__dict__; see
        # https://docs.python.org/3/howto/descriptor.html
        raise AttributeError("can't set info attribute.")

    def __bool__(self):
        return self.format is not None

    # PY2
    __nonzero__ = __bool__

    def __call__(self):
        """Create a dict with file information, including missing pieces."""
        info = {}
        if self:
            for attr in self.attr_names:
                value = getattr(self, attr)
                if value is not None:
                    info[attr] = value
            if self.missing:
                info['missing'] = self.missing

        return info

    def __repr__(self):
        # Use the repr for quick display of file information.
        if self._parent is None:
            return super(VLBIInfoBase, self).__repr__()

        if not self:
            return 'Not parsable. Wrong format?'

        result = ''
        for attr in self.attr_names:
            value = getattr(self, attr)
            if value is not None:
                if hasattr(value, 'isot'):
                    value.precision = 9
                    value = value.isot
                elif attr == 'sample_rate':
                    value = value.to(u.MHz)
                result += '{} = {}\n'.format(attr, value)

        if self.missing:
            result += '\n'
            prefix = 'missing: '
            for msg in sorted(set(self.missing.values())):
                keys = sorted(set(key for key in self.missing
                                  if self.missing[key] == msg))
                result += "{} {}: {}\n".format(prefix, ', '.join(keys), msg)
                prefix = ' ' * len(prefix)

        return result


class VLBIFileReaderInfo(VLBIInfoBase):
    """Standardized information on file readers.

    The ``info`` descriptor has a number of standard attributes, which are
    determined from arguments passed in opening the file, from the first header
    (``info.header0``) and from possibly scanning the file to determine the
    duration of frames.

    Attributes
    ----------
    format : str or `None`
        File format, or `None` if the underlying file cannot be parsed.
    frame_rate : `~astropy.units.Quantity`
        Number of data frames per unit of time.
    sample_rate : `~astropy.units.Quantity`
        Complete samples per unit of time.
    samples_per_frame : int
        Number of complete samples in each frame.
    sample_shape : tuple
        Dimensions of each complete sample (e.g., ``(nchan,)``).
    bps : int
        Number of bits used to encode each elementary sample.
    complex_data : bool
        Whether the data are complex.
    start_time : `~astropy.time.Time`
        Time of the first complete sample.
    missing : dict
        Entries are keyed by names of arguments that should be passed to
        the file reader to obtain full information. The associated entries
        explain why these arguments are needed.

    Examples
    --------
    The most common use is simply to print information::

        >>> from baseband.data import SAMPLE_MARK5B
        >>> from baseband import mark5b
        >>> fh = mark5b.open(SAMPLE_MARK5B, 'rb')
        >>> fh.info
        File information:
        format = mark5b
        frame_rate = 6400.0 Hz
        bps = 2
        complex_data = False
        <BLANKLINE>
        missing:  nchan: needed to determine sample shape and rate.
                  kday, ref_time: needed to infer full times.
        <BLANKLINE>
        >>> fh.close()

        >>> fh = mark5b.open(SAMPLE_MARK5B, 'rb', kday=56000, nchan=8)
        >>> fh.info
        File information:
        format = mark5b
        frame_rate = 6400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 5000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        start_time = 2014-06-13T05:30:01.000000000
        >>> fh.close()
    """
    attr_names = ('format', 'frame_rate', 'sample_rate', 'samples_per_frame',
                  'sample_shape', 'bps', 'complex_data', 'start_time')
    _header0_attrs = ('bps', 'complex_data', 'samples_per_frame',
                      'sample_shape')

    def _get_header0(self):
        fh = self._parent
        old_offset = fh.tell()
        try:
            fh.seek(0)
            return fh.read_header()
        except Exception:
            return None
        finally:
            fh.seek(old_offset)

    def _get_format(self):
        return self._parent.__class__.__name__.split('File')[0].lower()

    def _get_frame_rate(self):
        try:
            return self._parent.get_frame_rate()
        except Exception:
            return None

    def _get_start_time(self):
        try:
            return self.header0.time
        except Exception:
            return None

    def _collect_info(self):
        super(VLBIFileReaderInfo, self)._collect_info()
        self.header0 = self._get_header0()
        if self.header0 is not None:
            for attr in self._header0_attrs:
                setattr(self, attr, getattr(self.header0, attr))
            self.format = self._get_format()
            self.frame_rate = self._get_frame_rate()
            if (self.frame_rate is not None and
                    self.samples_per_frame is not None):
                self.sample_rate = self.frame_rate * self.samples_per_frame
            self.start_time = self._get_start_time()

    def __repr__(self):
        result = 'File information:\n'
        result += super(VLBIFileReaderInfo, self).__repr__()
        return result


class VLBIStreamReaderInfo(VLBIInfoBase):
    """Standardized information on stream readers.

    The ``info`` descriptor provides a few standard attributes, all of which
    can also be accessed directly on the stream filehandle. More detailed
    information on the underlying file is stored in its info, accessible via
    ``info.file_info``.

    Attributes
    ----------
    start_time : `~astropy.time.Time`
        Time of the first complete sample.
    stop_time : `~astropy.time.Time`
        Time of the complete sample just beyond the end of the file.
    sample_rate : `~astropy.units.Quantity`
        Complete samples per unit of time.
    shape : tuple
        Equivalent shape of the whole file, i.e., combining the number of
        complete samples and the shape of those samples.
    bps : int
        Number of bits used to encode each elementary sample.
    complex_data : bool
        Whether the data are complex.
    """
    attr_names = ('start_time', 'stop_time', 'sample_rate', 'shape',
                  'format', 'bps', 'complex_data')
    _parent_attrs = tuple(attr for attr in attr_names if attr != 'format')

    def _raw_file_info(self):
        # Mostly here so GSB can override.
        return self._parent.fh_raw.info

    def _collect_info(self):
        super(VLBIStreamReaderInfo, self)._collect_info()
        # We also want the raw info.
        self.file_info = self._raw_file_info()
        self.format = self.file_info.format

    def _up_to_date(self):
        # Stream readers cannot after initialization, so the check is easy.
        return True

    def __call__(self):
        """Create a dict with information about the stream and the raw file."""
        info = super(VLBIStreamReaderInfo, self).__call__()
        info['file_info'] = self.file_info()
        return info

    def __repr__(self):
        result = 'Stream information:\n'
        result += super(VLBIStreamReaderInfo, self).__repr__()
        file_info = getattr(self, 'file_info', None)
        if file_info is not None:
            # Add information from the raw file.
            raw_attrs = file_info.attr_names
            raw_only_attrs = [attr for attr in raw_attrs
                              if attr not in self.attr_names]
            try:
                file_info.attr_names = raw_only_attrs
                result += '\n' + repr(file_info)
            finally:
                file_info.attr_names = raw_attrs

        return result



# Licensed under the GPLv3 - see LICENSE
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import pytest
import numpy as np
import astropy.units as u
from astropy.time import Time
from astropy.tests.helper import catch_warnings
from ... import mark4
from ...vlbi_base.encoding import OPTIMAL_2BIT_HIGH
from ..header import Mark4TrackHeader
from ..payload import reorder32, reorder64
from ...data import (SAMPLE_MARK4 as SAMPLE_FILE,
                     SAMPLE_MARK4_32TRACK as SAMPLE_32TRACK,
                     SAMPLE_MARK4_32TRACK_FANOUT2 as SAMPLE_32TRACK_FANOUT2,
                     SAMPLE_MARK4_16TRACK as SAMPLE_16TRACK)

# Results from mark5access on 2015-JAN-22.
# m5d evn/Ar/gp052d_ar_no0021 MKIV1_4-512-8-2 1000
# Mark5 stream: 0x1a54140
#   stream = File-1/1=evn/gp052d_ar_no0021
#   format = MKIV1_4-512-8-2/1 = 1
#   start mjd/sec = 53171 27492.475000000
#   frame duration = 2500000.00 ns
#   framenum = 0
#   sample rate = 32000000 Hz
#   offset = 3208
#   framebytes = 160000 bytes
#   datasize = 160000 bytes
#   sample granularity = 4
#   frame granularity = 1
#   gframens = 2500000
#   payload offset = -512
#   read position = 0
#   data window size = 1048576 bytes
#   ...
# data at 17, nonzero at line 657 -> item 640.
# Initially this seemed strange, since PAYLOAD_NBITS=20000 leads to 80000
# elements, so one would have expected VALIDSTART*4=96*4=384.
# But the mark5 code has PAYLOAD_OFFSET=(VALIDEND-20000)*f->ntrack/8 = 64*8
# Since each sample takes 2 bytes, one thus expects 384+64*8/2=640. OK.
# So, lines 639--641:
#  0  0  0  0  0  0  0  0
# -1  1  1 -3 -3 -3  1 -1
#  1  1 -3  1  1 -3 -1 -1


class TestMark4(object):
    def test_header_stream(self):
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            stream = np.fromfile(fh, dtype='<u8', count=5 * 32)
        # check sync words in right place
        assert np.all(stream[64:80] == 0xffffffffffffffff)
        assert mark4.header.crc12.check(stream)
        assert np.all(mark4.header.crc12(stream[:-12]) == stream[-12:])
        words = mark4.header.stream2words(stream)
        assert np.all(mark4.header.words2stream(words) == stream)

    def test_header(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            fh.seek(-10, 2)
            with pytest.raises(EOFError):
                mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)

        assert len(header) == 64
        assert header.track_id[0] == 2 and header.track_id[-1] == 33
        assert header.nbytes == 160 * 64 // 8
        assert header.fanout == 4
        assert header.bps == 2
        assert not np.all(~header['magnitude_bit'])
        assert header.nchan == 8
        assert int(header.time.mjd) == 56824
        assert header.time.isot == '2014-06-16T07:38:12.47500'
        assert header.samples_per_frame == 20000 * 4
        assert header.frame_nbytes == 20000 * 64 // 8
        assert header.payload_nbytes == header.frame_nbytes - header.nbytes
        assert header.mutable is False
        assert header.nsb == 1
        assert np.all(header.converters['converter'] ==
                      [0, 2, 1, 3, 4, 6, 5, 7])
        assert np.all(header.converters['lsb'])
        assert repr(header).startswith('<Mark4Header bcd_headstack1: [0')
        with open(str(tmpdir.join('test.m4')), 'w+b') as s:
            header.tofile(s)
            s.seek(0)
            header2 = mark4.Mark4Header.fromfile(s, header.ntrack,
                                                 header.decade)
        assert header2 == header
        assert header2.mutable is False
        header3 = mark4.Mark4Header.fromkeys(header.ntrack, header.decade,
                                             **header)
        assert header3 == header
        assert header3.mutable is True
        # Try initialising with properties instead of keywords.
        # Here, we let time imply the decade, bcd_unit_year, bcd_day, bcd_hour,
        # bcd_minute, bcd_second, bcd_fraction;
        # and ntrack, samples_per_frame, bps define headstack_id, bcd_track_id,
        # fan_out, magnitude_bit, and converter_id.
        header4 = mark4.Mark4Header.fromvalues(
            ntrack=64, samples_per_frame=80000, bps=2, nsb=1, time=header.time,
            system_id=108)
        assert header4 == header
        assert header4.mutable is True
        # Check that passing a year into decade leads to an error.
        with pytest.raises(AssertionError):
            mark4.Mark4Header(header.words, decade=2014)
        # Check that passing approximate ref_time is equivalent to passing a
        # decade.
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header5 = mark4.Mark4Header.fromfile(
                fh, ntrack=64, ref_time=Time('2010:351:12:00:00.0'))
            assert header5 == header
            fh.seek(0xa88)
            header6 = mark4.Mark4Header.fromfile(
                fh, ntrack=64, ref_time=Time('2018:113:16:30:00.0'))
            assert header6 == header
        # Check changing properties.
        header7 = header.copy()
        assert header7 == header
        assert header7.mutable is True
        header7['bcd_headstack1'] = 0x5566
        assert np.all(header7['bcd_headstack1'] == 0x5566)
        assert header7 != header
        header7['bcd_headstack1'] = np.hstack((0x7788,
                                               header7['bcd_headstack1'][1:]))
        assert header7['bcd_headstack1'][0] == 0x7788
        assert np.all(header7['bcd_headstack1'][1:] == 0x5566)
        # Check it doesn't work on non-mutable header
        with pytest.raises(TypeError):
            header['bcd_headstack1'] = 0
        # Check time assignment
        with pytest.raises(ValueError):
            header7.time = header.time + 0.1 * u.ms
        header7.bps = 1
        assert np.all(~header7['magnitude_bit'])
        header7.bps = 2
        assert not np.all(~header7['magnitude_bit'])
        with pytest.raises(ValueError):
            header7.bps = 4
        with pytest.raises(AttributeError):
            header7.ntrack = 51
        with pytest.raises(AttributeError):
            header7.frame_nbytes = header.frame_nbytes
        with pytest.raises(AttributeError):
            header7.payload_nbytes = header.payload_nbytes
        header7.nchan = 16
        assert header7.nchan == 16 and header7.bps == 1
        # OK, this is silly, but why not...
        header7.time = header.time + np.arange(64) * 125 * u.ms
        assert len(header7.time) == 64
        assert np.all(abs(header7.time - header.time -
                          np.arange(64) * 125 * u.ms) < 1.*u.ns)
        with pytest.raises(ValueError):  # different decades
            header7.time = header.time - (np.linspace(0, 20, 64) // 4) * 4*u.yr
        # Check slicing.
        header8 = header[:2]
        assert type(header8) is mark4.Mark4Header
        assert len(header8) == 2
        header9 = header[10]
        assert type(header9) is Mark4TrackHeader
        assert repr(header9).startswith('<Mark4TrackHeader bcd_headstack1: 0')
        assert repr(header[:1]).startswith('<Mark4Header bcd_headstack1: 0')
        header10 = Mark4TrackHeader(header9.words, decade=2010)
        assert header10 == header9
        header11 = Mark4TrackHeader.fromvalues(decade=2010, **header9)
        assert header11 == header9
        # Check passing decade=None still reads the header, and we can set
        # decade afterward.
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header12 = mark4.Mark4Header.fromfile(fh, ntrack=64)
            assert header12.decade is None
            header12.decade = 2010
            assert header12 == header
        with pytest.raises(IndexError):
            header[65]
        with pytest.raises(ValueError):
            header[np.array([[0, 1], [2, 3]])]
        # Check that one can construct crazy headers, even if not much works.
        header13 = mark4.Mark4Header(None, ntrack=53, decade=2010,
                                     verify=False)
        header13.time = header.time
        assert header13.ntrack == 53
        assert abs(header13.time - header.time) < 1. * u.ns

    @pytest.mark.parametrize(('unit_year', 'ref_time', 'decade'),
                             [(5, Time('2014:1:12:00:00'), 2010),
                              (5, Time('2009:362:19:27:33'), 2000),
                              (4, Time('2009:001:19:27:33'), 2010),
                              (3, Time('2018:117:6:42:15'), 2020),
                              (4, Time('2018:117:6:42:15'), 2010)])
    def test_infer_decade(self, unit_year, ref_time, decade):
        # Check that infer_decade returns proper decade for
        # ref_time.year - 5 < year < ref_time.year + 5, and uses bankers'
        # rounding at the boundaries.
        header = mark4.header.Mark4Header(None, ntrack=16, verify=False)
        header['bcd_unit_year'] = unit_year
        header.infer_decade(ref_time)
        assert header.decade == decade

    def test_decoding(self):
        """Check that look-up levels are consistent with mark5access."""
        o2h = OPTIMAL_2BIT_HIGH
        assert np.all(mark4.payload.lut1bit[0] == 1.)
        assert np.all(mark4.payload.lut1bit[0xff] == -1.)
        assert np.all(mark4.payload.lut1bit.astype(int) ==
                      1 - 2 * ((np.arange(256)[:, np.newaxis] >>
                                np.arange(8)) & 1))
        assert np.all(mark4.payload.lut2bit1[0] == -o2h)
        assert np.all(mark4.payload.lut2bit1[0x55] == 1.)
        assert np.all(mark4.payload.lut2bit1[0xaa] == -1.)
        assert np.all(mark4.payload.lut2bit1[0xff] == o2h)
        assert np.all(mark4.payload.lut2bit2[0] == -o2h)
        assert np.all(mark4.payload.lut2bit2[0xcc] == -1.)
        assert np.all(mark4.payload.lut2bit2[0x33] == 1.)
        assert np.all(mark4.payload.lut2bit2[0xff] == o2h)
        assert np.all(mark4.payload.lut2bit3[0] == -o2h)
        assert np.all(mark4.payload.lut2bit3[0xf0] == -1.)
        assert np.all(mark4.payload.lut2bit3[0x0f] == 1.)
        assert np.all(mark4.payload.lut2bit3[0xff] == o2h)

    def test_payload_reorder(self):
        """Test that bit reordering is consistent with mark5access."""
        check = np.array([738811025863578102], dtype='<u8').view('u8')
        expected = np.array([118, 209, 53, 244, 148, 217, 64, 10])
        assert np.all(reorder64(check).view(np.uint8) == expected)
        assert np.all(reorder32(check.view('u4')).view(np.uint8) ==
                      expected)

    def test_payload(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            payload = mark4.Mark4Payload.fromfile(fh, header)
        assert payload.nbytes == (20000 - 160) * 64 // 8
        assert len(payload) == (20000 - 160) * 4
        assert payload.shape == ((20000 - 160) * 4, 8)
        assert payload.size == 634880
        assert payload.ndim == 2
        # Check sample shape validity.
        assert payload.sample_shape == (8,)
        assert payload.sample_shape.nchan == 8
        assert payload.dtype == np.float32
        assert np.all(payload[0].astype(int) ==
                      np.array([-1, +1, +1, -3, -3, -3, +1, -1]))
        assert np.all(payload[1].astype(int) ==
                      np.array([+1, +1, -3, +1, +1, -3, -1, -1]))

        with open(str(tmpdir.join('test.m4')), 'w+b') as s:
            payload.tofile(s)
            s.seek(0)
            payload2 = mark4.Mark4Payload.fromfile(s, header)
            assert payload2 == payload
            with pytest.raises(EOFError):
                # Too few bytes.
                s.seek(100)
                mark4.Mark4Payload.fromfile(s, header)
        payload3 = mark4.Mark4Payload.fromdata(payload.data, header)
        assert payload3 == payload
        payload4 = mark4.Mark4Payload(payload.words, nchan=8, bps=2, fanout=4)
        assert payload4 == payload
        with pytest.raises(ValueError):
            # Wrong number of channels.
            mark4.Mark4Payload.fromdata(np.empty((payload.shape[0], 2)),
                                        header)
        with pytest.raises(ValueError):
            # Too few data.
            mark4.Mark4Payload.fromdata(payload[:100], header)

        with pytest.raises(ValueError):
            # Wrong data type
            mark4.Mark4Payload.fromdata(np.zeros((5000, 8), np.complex64),
                                        header)
        with pytest.raises(ValueError):
            # Wrong encoded data type for implied number of tracks of 32.
            mark4.Mark4Payload(payload.words, nchan=4, bps=2, fanout=4)
        with pytest.raises(ValueError):
            # Not little-endian encoded data.
            mark4.Mark4Payload(payload.words.astype('>u8'), header)
        with pytest.raises(ValueError):
            # Wrong number of tracks in encoded data.
            mark4.Mark4Payload(payload.words.view('<u4'), header)

    @pytest.mark.parametrize('item', (2, (), -1, slice(1, 3), slice(2, 4),
                                      slice(2, 4), slice(-3, None),
                                      (2, slice(3, 5)), (10, 4),
                                      (slice(None), 5)))
    def test_payload_getitem_setitem(self, item):
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            payload = mark4.Mark4Payload.fromfile(fh, header)
        sel_data = payload.data[item]
        assert np.all(payload[item] == sel_data)
        payload2 = mark4.Mark4Payload(payload.words.copy(), header)
        assert payload2 == payload
        payload2[item] = -sel_data
        check = payload.data
        check[item] = -sel_data
        assert np.all(payload2[item] == -sel_data)
        assert np.all(payload2.data == check)
        assert payload2 != payload
        payload2[item] = sel_data
        assert np.all(payload2[item] == sel_data)
        assert payload2 == payload

    def test_binary_file_reader(self):
        with mark4.open(SAMPLE_FILE, 'rb', decade=2010, ntrack=64) as fh:
            fh.locate_frame()
            assert fh.tell() == 0xa88
            header = mark4.Mark4Header.fromfile(fh, decade=2010, ntrack=64)
            fh.seek(0xa88)
            header2 = fh.read_header()
            current_pos = fh.tell()
            assert header2 == header
            frame_rate = fh.get_frame_rate()
            assert abs(frame_rate -
                       32 * u.MHz / header.samples_per_frame) < 1 * u.nHz
            assert fh.tell() == current_pos
            repr_fh = repr(fh)

        assert repr_fh.startswith('Mark4FileReader')
        assert 'ntrack=64, decade=2010, ref_time=None' in repr_fh

    def test_frame(self, tmpdir):
        with mark4.open(SAMPLE_FILE, 'rb', decade=2010, ntrack=64) as fh:
            fh.seek(0xa88)
            header = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            payload = mark4.Mark4Payload.fromfile(fh, header)
            fh.seek(0xa88)
            frame = fh.read_frame()

        assert frame.header == header
        assert frame.payload == payload
        assert frame.valid is True
        assert len(frame) == len(payload) + 640
        assert frame.sample_shape == payload.sample_shape
        assert frame.shape == (len(frame),) + frame.sample_shape
        assert frame.size == len(frame) * np.prod(frame.sample_shape)
        assert frame.ndim == payload.ndim
        assert frame == mark4.Mark4Frame(header, payload)
        data = frame.data
        assert np.all(data[:640] == 0.)
        assert np.all(data[640].astype(int) ==
                      np.array([-1, +1, +1, -3, -3, -3, +1, -1]))
        assert np.all(data[641].astype(int) ==
                      np.array([+1, +1, -3, +1, +1, -3, -1, -1]))
        # Check writing and round-trip.
        with open(str(tmpdir.join('test.m4')), 'w+b') as s:
            frame.tofile(s)
            s.seek(0)
            frame2 = mark4.Mark4Frame.fromfile(s, ntrack=64, decade=2010)
        assert frame2 == frame
        # Need frame.data here, since payload.data does not contain pieces
        # overwritten by header.
        frame3 = mark4.Mark4Frame.fromdata(frame.data, header)
        assert frame3 == frame
        frame4 = mark4.Mark4Frame.fromdata(frame.data, ntrack=64,
                                           decade=2010, **header)
        assert frame4 == frame
        header5 = header.copy()
        frame5 = mark4.Mark4Frame(header5, payload, valid=False)
        assert frame5.valid is False
        assert np.all(frame5.data == 0.)
        frame5.valid = True
        assert frame5 == frame
        frame5.valid = False
        assert np.all(frame5.data == 0.)

        # Check passing in a reference time.
        with mark4.open(SAMPLE_FILE, 'rb', ref_time=Time('2009:345:15:00:00'),
                        ntrack=64) as fh:
            fh.seek(0xa88)
            frame6 = fh.read_frame()
            assert frame6 == frame

        with mark4.open(SAMPLE_FILE, 'rb', ref_time=Time('2019:1:9:00:00'),
                        ntrack=64) as fh:
            fh.seek(0xa88)
            frame7 = fh.read_frame()
            assert frame7 == frame

    def test_frame_getitem_setitem(self):
        with mark4.open(SAMPLE_FILE, 'rb', ref_time=Time('2009:345:15:00:00'),
                        ntrack=64) as fh:
            fh.seek(0xa88)
            frame = fh.read_frame()
            header = frame.header
            data = frame.data
        # Check __getitem__.
        assert np.all(frame['magnitude_bit'] == header['magnitude_bit'])
        # Only invalid values (compare with data to ensure shape is right).
        assert np.all(frame[10:90] == data[10:90])
        assert np.all(frame[10:90:5] == data[10:90:5])
        assert np.all(frame[10:90:5, :4] == data[10:90:5, :4])
        # Mixed invalid, valid.
        assert np.all(frame[635:655] == data[635:655])
        # Check we do stepping correctly around the boundary.
        for start in range(634, 642):
            assert np.all(frame[start:655:5] == data[start:655:5])
        assert np.all(frame[635:655:5, 5] == data[635:655:5, 5])
        # All valid.
        assert np.all(frame[935:955] == data[935:955])
        assert np.all(frame[935:955:5] == data[935:955:5])
        assert np.all(frame[935:955:5, 5] == data[935:955:5, 5])
        # Check scalar, including whether the shape is correct.
        assert frame[935].shape == data[935].shape
        assert np.all(frame[639] == data[639])
        assert np.all(frame[640] == data[640])
        # Check __setitem__.
        # Default frame read from a file is not mutable.
        with pytest.raises(ValueError):
            frame[640:] = 0.
        # So create a mutable version of the frame.
        frame = mark4.Mark4Frame.fromdata(frame.data, header.copy())
        # Check that setting invalid part has no effect.
        frame[10:90] == 1.
        frame[10:90:5] == 1.
        frame[10:90:5, :4] == 1.
        assert np.all(frame[:640] == 0.)
        # Mixed invalid, valid.
        frame[635:655] = 1.
        assert np.all(frame[635:640] == 0.)
        assert np.all(frame[640:655] == 1.)
        frame[635:655] = data[635:655]
        assert np.all(frame[:] == data)
        # Check we do stepping correctly around the boundary.
        for start in range(634, 642):
            frame[start:655:5] = 1.
            valid_start = 640 + start % 5
            if start < 640:
                assert np.all(frame[start:640:5] == 0.)
            assert np.all(frame[valid_start:655:5] == 1.)
        frame[634:655] = data[634:655]
        frame[635:655:5, 5] = -data[635:655:5, 5]
        assert np.all(frame[635:655:5, 5] == -data[635:655:5, 5])
        assert np.all(frame[635:655:5, :5] == data[635:655:5, :5])
        assert np.all(frame[635:655:5, 6:] == data[635:655:5, 6:])
        # All valid.
        frame[935:955] = 1.
        assert np.all(frame[935:955] == 1.)
        frame[935:955:5] = -1.
        assert np.all(frame[935:955:5] == -1.)
        assert np.all(frame[936:955:5] == 1.)
        frame[935:955:5, 5] = 1.
        assert np.all(frame[935:955:5, 5] == 1.)
        assert np.all(frame[935:955:5, :5] == -1.)
        frame[935:955] = data[935:955]
        # Check scalar, including whether the shape is correct.
        frame[935] = -data[935]
        assert np.all(frame[935] == -data[935])
        assert np.all(frame[934] == data[934])
        assert np.all(frame[936] == data[936])
        frame[:] = data
        assert np.all(frame[:] == data)
        # Finally, check that __getitem__ always returns 0 for invalid data.
        frame.valid = False
        assert frame[655, 0] == 0.
        assert np.all(frame[930:950] == 0.)
        assert np.all(frame[630:650:5, :4] == 0.)
        frame.valid = True
        # And check propagation of strings to header.
        frame['bcd_headstack2'] = 0
        assert np.all(frame.header['bcd_headstack2'] == 0.)

    def test_header_times(self):
        with mark4.open(SAMPLE_FILE, 'rb', decade=2010, ntrack=64) as fh:
            fh.seek(0xa88)
            header0 = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            start_time = header0.time
            # Use frame size, since header adds to payload.
            samples_per_frame = header0.frame_nbytes * 8 // 2 // 8
            frame_rate = 32. * u.MHz / samples_per_frame
            frame_duration = 1. / frame_rate
            fh.seek(0xa88)
            for frame_nr in range(100):
                try:
                    frame = fh.read_frame()
                except EOFError:
                    break
                header_time = frame.header.time
                expected = start_time + frame_nr * frame_duration
                assert abs(header_time - expected) < 1. * u.ns

    def test_find_header(self, tmpdir):
        # Below, the tests set the file pointer to very close to a header,
        # since otherwise they run *very* slow.  This is somehow related to
        # pytest, since speed is not a big issue running stuff on its own.
        with mark4.open(SAMPLE_FILE, 'rb', decade=2010) as fh:
            fh.seek(0xa88)
            header0 = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)
            fh.seek(0)
            header_0 = fh.find_header()
            assert fh.tell() == 0xa88
            assert fh.ntrack == 64
            assert header_0 == header0
            fh.seek(0xa89)
            header_0xa89 = fh.find_header()
            assert fh.tell() == 0xa88 + header0.frame_nbytes
            fh.seek(160000)
            header_160000f = fh.find_header(forward=True)
            assert fh.tell() == 0xa88 + header0.frame_nbytes
            fh.seek(0xa87)
            header_0xa87b = fh.find_header(forward=False)
            assert header_0xa87b is None
            assert fh.tell() == 0xa87
            fh.seek(0xa88)
            header_0xa88f = fh.find_header()
            assert fh.tell() == 0xa88
            fh.seek(0xa88)
            header_0xa88b = fh.find_header(forward=False)
            assert fh.tell() == 0xa88
            fh.seek(0xa88 + 100)
            header_100b = fh.find_header(forward=False)
            assert fh.tell() == 0xa88
            fh.seek(-10000, 2)
            header_m10000b = fh.find_header(forward=False)
            assert fh.tell() == 0xa88 + 2*header0.frame_nbytes
            fh.seek(-300, 2)
            header_end = fh.find_header(forward=True)
            assert header_end is None
        assert header_100b == header_0
        assert header_0xa88f == header_0
        assert header_0xa88b == header_0
        assert header_0xa89 == header_160000f
        assert abs(header_160000f.time - header_0.time - 2.5*u.ms) < 1.*u.ns
        assert abs(header_m10000b.time - header_0.time - 5*u.ms) < 1.*u.ns
        # test small file
        with open(SAMPLE_FILE, 'rb') as fh:
            # One that simply is too small altogether
            m4_test = str(tmpdir.join('test.m4'))
            with open(m4_test, 'w+b') as s, mark4.open(s, 'rb',
                                                       ntrack=64) as fh_short:
                s.write(fh.read(80000))
                fh_short.seek(100)
                assert fh_short.find_header() is None
                assert fh_short.tell() == 100
                assert fh_short.find_header(forward=False) is None
                assert fh_short.tell() == 100

            # And one that could fit one frame, but doesn't.
            with open(m4_test, 'w+b') as s, mark4.open(s, 'rb',
                                                       ntrack=64) as fh_short:
                fh.seek(0)
                s.write(fh.read(162690))
                fh_short.seek(200)
                assert fh_short.find_header() is None
                assert fh_short.tell() == 200
                assert fh_short.find_header(forward=False) is None
                assert fh_short.tell() == 200

            # now add enough that the file does include a complete header.
            with open(m4_test, 'w+b') as s, mark4.open(s, 'rb',
                                                       ntrack=64) as fh_short2:
                fh.seek(0)
                s.write(fh.read(163000))
                s.seek(0)
                fh_short2.seek(100)
                header_100f = fh_short2.find_header()
                assert fh_short2.tell() == 0xa88
                fh_short2.seek(-1000, 2)
                header_m1000b = fh_short2.find_header(forward=False)
                assert fh_short2.tell() == 0xa88
            assert header_100f == header0
            assert header_m1000b == header0

    def test_determine_ntrack(self):
        with mark4.open(SAMPLE_FILE, 'rb', ntrack=64) as fh:
            offset0 = fh.locate_frame()
            assert offset0 == 2696
        with mark4.open(SAMPLE_FILE, 'rb') as fh:
            assert fh.ntrack is None
            ntrack = fh.determine_ntrack()
            assert ntrack == fh.ntrack == 64
            assert fh.fh_raw.tell() == offset0

        with mark4.open(SAMPLE_32TRACK, 'rb', ntrack=32) as fh:
            # Seek past first frame header; find second frame.
            fh.seek(10000)
            offset0 = fh.locate_frame()
            assert offset0 == 89656
        with mark4.open(SAMPLE_32TRACK, 'rb') as fh:
            fh.seek(10000)
            ntrack = fh.determine_ntrack()
            assert fh.fh_raw.tell() == offset0
            assert ntrack == fh.ntrack == 32

        with mark4.open(SAMPLE_32TRACK_FANOUT2, 'rb', ntrack=32) as fh:
            offset0 = fh.locate_frame()
            assert offset0 == 17436
        with mark4.open(SAMPLE_32TRACK_FANOUT2, 'rb') as fh:
            ntrack = fh.determine_ntrack()
            assert fh.fh_raw.tell() == offset0
            assert ntrack == fh.ntrack == 32
            assert fh.ntrack == 32

    def test_filestreamer(self, tmpdir):
        with mark4.open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(0xa88)
            header = mark4.Mark4Header.fromfile(fh, ntrack=64, decade=2010)

        with mark4.open(SAMPLE_FILE, 'rs', sample_rate=32*u.MHz,
                        ntrack=64, decade=2010) as fh:
            assert header == fh.header0
            assert fh.samples_per_frame == 80000
            assert fh.sample_shape == (8,)
            assert fh.shape == (2 * fh.samples_per_frame,) + fh.sample_shape
            assert fh.size == np.prod(fh.shape)
            assert fh.ndim == len(fh.shape)
            assert fh.sample_rate == 32 * u.MHz
            record = fh.read(642)
            assert fh.tell() == 642
            # regression test against #4, of incorrect frame offsets.
            fh.seek(80000 + 639)
            record2 = fh.read(2)
            assert fh.tell() == 80641
            # Raw file should be just after frame 1.
            assert fh.fh_raw.tell() == 0xa88 + 2 * fh.header0.frame_nbytes
            # Test seeker works with both int and str values for whence
            assert fh.seek(13, 0) == fh.seek(13, 'start')
            assert fh.seek(-13, 2) == fh.seek(-13, 'end')
            fhseek_int = fh.seek(17, 1)
            fh.seek(-17, 'current')
            fhseek_str = fh.seek(17, 'current')
            assert fhseek_int == fhseek_str
            with pytest.raises(ValueError):
                fh.seek(0, 'last')
            fh.seek(1, 'end')
            with pytest.raises(EOFError):
                fh.read()

        assert record.shape == (642, 8)
        assert np.all(record[:640] == 0.)
        assert np.all(record.astype(int)[640] ==
                      np.array([-1, +1, +1, -3, -3, -3, +1, -1]))
        assert np.all(record.astype(int)[641] ==
                      np.array([+1, +1, -3, +1, +1, -3, -1, -1]))
        assert record2.shape == (2, 8)
        assert np.all(record2[0] == 0.)
        assert not np.any(record2[1] == 0.)

        # Check passing a time object into decade.
        with mark4.open(SAMPLE_FILE, 'rs', ntrack=64,
                        ref_time=Time('2018:364:23:59:59')) as fh:
            assert header == fh.header0
        with mark4.open(SAMPLE_FILE, 'rs', ntrack=64,
                        ref_time=Time(56039.5, format='mjd')) as fh:
            assert header == fh.header0

        # Test if _get_frame_rate automatic frame rate calculator works,
        # returns same header and payload info.
        with mark4.open(SAMPLE_FILE, 'rs', ntrack=64, decade=2010) as fh:
            assert header == fh.header0
            assert fh.samples_per_frame == 80000
            assert fh.sample_rate == 32 * u.MHz
            record3 = fh.read(642)

        assert np.all(record3 == record)

        # Test if automatic ntrack and frame rate detectors work together.
        with mark4.open(SAMPLE_FILE, 'rs', decade=2010) as fh:
            assert header == fh.header0
            assert fh.sample_rate == 32 * u.MHz
            fh.seek(80000 + 639)
            record4 = fh.read(2)

        assert np.all(record4 == record2)

        with mark4.open(SAMPLE_FILE, 'rs', sample_rate=32*u.MHz,
                        ntrack=64, decade=2010) as fh:
            start_time = fh.time
            record = fh.read()
            fh_raw_tell1 = fh.fh_raw.tell()
            stop_time = fh.time

        rewritten_file = str(tmpdir.join('rewritten.m4'))
        with mark4.open(rewritten_file, 'ws', sample_rate=32*u.MHz,
                        time=start_time, ntrack=64, bps=2, fanout=4) as fw:
            assert fw.sample_rate == 32 * u.MHz
            # write in bits and pieces and with some invalid data as well.
            fw.write(record[:11])
            fw.write(record[11:80000])
            fw.write(record[80000:], valid=False)
            assert fw.tell(unit='time') == stop_time

        with mark4.open(rewritten_file, 'rs', sample_rate=32*u.MHz,
                        ntrack=64, decade=2010, subset=[3, 4]) as fh:
            assert fh.time == start_time
            assert fh.time == fh.tell(unit='time')
            assert fh.sample_rate == 32 * u.MHz
            record5 = fh.read(160000)
            assert fh.time == stop_time
            assert fh.sample_shape == (2,)
            assert np.all(record5[:80000] == record[:80000, 3:5])
            assert np.all(record5[80000:] == 0.)

        # Check files can be made byte-for-byte identical.  Here, we use the
        # original header so we set stuff like head_stack, etc.
        with open(str(tmpdir.join('test.m4')), 'w+b') as s, \
                mark4.open(s, 'ws', header0=header,
                           sample_rate=32*u.MHz) as fw:
            fw.write(record)
            number_of_bytes = s.tell()
            assert number_of_bytes == fh_raw_tell1 - 0xa88

            s.seek(0)
            with open(SAMPLE_FILE, 'rb') as fr:
                fr.seek(0xa88)
                orig_bytes = fr.read(number_of_bytes)
                conv_bytes = s.read()
                assert conv_bytes == orig_bytes

        # Test that squeeze attribute works on read (including in-place read).
        with mark4.open(SAMPLE_FILE, 'rs', ntrack=64, decade=2010,
                        subset=0) as fh:
            assert fh.sample_shape == ()
            assert fh.read(1).shape == (1,)
            fh.seek(0)
            out = np.zeros(12)
            fh.read(out=out)
            assert fh.tell() == 12
            assert np.all(out == record[:12, 0])

        with mark4.open(SAMPLE_FILE, 'rs', ntrack=64, decade=2010,
                        subset=[0], squeeze=False) as fh:
            assert fh.subset == ([0],)
            assert fh.sample_shape == (1,)
            assert fh.sample_shape.nchan == 1
            assert fh.read(1).shape == (1, 1)
            fh.seek(0)
            out = np.zeros((12, 1))
            fh.read(out=out)
            assert fh.tell() == 12
            assert np.all(out.squeeze() == record[:12, 0])

        # Test writing across decades.
        start_time = Time('2019:365:23:59:59.9975', precision=9)
        decadal_file = str(tmpdir.join('decade.m4'))
        with mark4.open(decadal_file, 'ws', sample_rate=32*u.MHz,
                        time=start_time, ntrack=64, bps=2, fanout=4) as fw:
            # write in bits and pieces and with some invalid data as well.
            fw.write(record)

        with mark4.open(decadal_file, 'rs', sample_rate=32*u.MHz,
                        ntrack=64, decade=2010) as fh:
            assert abs(fh.start_time - start_time) < 1. * u.ns
            assert fh.header0.decade == 2010
            record6 = fh.read()
            assert np.all(record6 == record)
            assert (abs(fh.time - Time('2020:1:00:00:00.0025', precision=9)) <
                    1. * u.ns)
            assert fh._frame.header.decade == 2020

    # Test that writing an incomplete stream is possible, and that frame set is
    # appropriately marked as invalid.
    @pytest.mark.parametrize('fill_value', (0., -999.))
    def test_incomplete_stream(self, tmpdir, fill_value):
        m4_incomplete = str(tmpdir.join('incomplete.m4'))
        with catch_warnings(UserWarning) as w:
            with mark4.open(SAMPLE_FILE, 'rs', ntrack=64, decade=2010) as fr:
                record = fr.read(10)
                with mark4.open(m4_incomplete, 'ws', header0=fr.header0,
                                sample_rate=32*u.MHz,
                                ntrack=64, decade=2010) as fw:
                    fw.write(record)
        assert len(w) == 1
        assert 'partial buffer' in str(w[0].message)
        with mark4.open(m4_incomplete, 'rs', sample_rate=32*u.MHz,
                        ntrack=64, decade=2010, fill_value=fill_value) as fwr:
            assert np.all(fwr.read() == fill_value)
            assert fwr.fill_value == fill_value

    def test_corrupt_stream(self, tmpdir):
        with mark4.open(SAMPLE_FILE, 'rb', decade=2010, ntrack=64) as fh, \
                open(str(tmpdir.join('test.m4')), 'w+b') as s:
            fh.seek(0xa88)
            frame = fh.read_frame()
            # Write single frame to file.
            frame.tofile(s)
            # Now add lots of data without headers.
            for i in range(5):
                frame.payload.tofile(s)
            s.seek(0)
            # With too many payload samples for one frame, f2.locate_frame
            # will fail.
            with pytest.raises(AssertionError):
                f2 = mark4.open(s, 'rs', sample_rate=32*u.MHz,
                                ntrack=64, decade=2010)

        with mark4.open(SAMPLE_FILE, 'rb', decade=2010, ntrack=64) as fh, \
                open(str(tmpdir.join('test.m4')), 'w+b') as s:
            fh.seek(0xa88)
            frame0 = fh.read_frame()
            frame1 = fh.read_frame()
            frame0.tofile(s)
            frame1.tofile(s)
            # now add lots of data without headers.
            for i in range(15):
                frame1.payload.tofile(s)
            s.seek(0)
            with mark4.open(s, 'rs', sample_rate=32*u.MHz,
                            ntrack=64, decade=2010) as f2:
                assert f2.header0 == frame0.header
                with pytest.raises(ValueError):
                    f2._last_header

    def test_stream_invalid(self):
        with pytest.raises(ValueError):
            mark4.open('ts.dat', 's')

    def test_stream_missing_decade(self):
        with pytest.raises(TypeError):
            mark4.open(SAMPLE_FILE, 'rs', ntrack=64)


class Test32TrackFanout4():
    def test_locate_frame(self):
        with mark4.open(SAMPLE_32TRACK, 'rb') as fh:
            assert fh.locate_frame() == 9656
            assert fh.ntrack == 32

    def test_header(self):
        with open(SAMPLE_32TRACK, 'rb') as fh:
            fh.seek(9656)
            header = mark4.Mark4Header.fromfile(fh, ntrack=32, decade=2010)

        # Try initialising with properties instead of keywords.
        # Here, we let
        # * time imply the decade, bcd_unit_year, bcd_day, bcd_hour,
        #   bcd_minute, bcd_second, bcd_fraction;
        # * ntrack, samples_per_frame, bps define headstack_id, bcd_track_id,
        #   fan_out, and magnitude_bit;
        # * nsb defines lsb_output and converter_id.
        header1 = mark4.Mark4Header.fromvalues(
            ntrack=32, samples_per_frame=80000, bps=2, nsb=2, time=header.time,
            system_id=108)
        assert header1 == header

    def test_file_streamer(self, tmpdir):
        with mark4.open(SAMPLE_32TRACK, 'rs', sample_rate=32*u.MHz,
                        ntrack=32, decade=2010) as fh:
            header0 = fh.header0
            assert fh.samples_per_frame == 80000
            assert fh.sample_rate == 32 * u.MHz
            start_time = fh.start_time
            assert start_time.yday == '2015:011:01:23:10.48500'
            record = fh.read(160000)
            fh_raw_tell1 = fh.fh_raw.tell()
            assert fh_raw_tell1 == 169656

        assert np.all(record[:640] == 0.)
        # Data retrieved using: m5d ar/rg10a_ar_no0014 MKIV1_4-256-4-2 700
        assert np.all(record[640:644].astype(int) == np.array(
            [[-1, 3, -1, -3],
             [3, 3, -3, 1],
             [-3, -1, 1, -1],
             [1, 3, 1, 3]]))

        fl = str(tmpdir.join('test.m4'))
        with mark4.open(fl, 'ws', header0=header0, sample_rate=32*u.MHz) as fw:
            fw.write(record)
            number_of_bytes = fw.fh_raw.tell()
            assert number_of_bytes == fh_raw_tell1 - 9656

        # Note: this test would not work if we wrote only a single record.
        with mark4.open(fl, 'rs', sample_rate=32*u.MHz,
                        ntrack=32, decade=2010) as fh:
            assert fh.start_time == start_time
            record2 = fh.read(1000)
            assert np.all(record2 == record[:1000])

        with open(fl, 'rb') as fh, open(SAMPLE_32TRACK, 'rb') as fr:
            fr.seek(9656)
            orig_bytes = fr.read(number_of_bytes)
            conv_bytes = fh.read()
            assert conv_bytes == orig_bytes


class Test32TrackFanout2():
    def test_locate_frame(self):
        with mark4.open(SAMPLE_32TRACK_FANOUT2, 'rb') as fh:
            assert fh.locate_frame() == 17436
            assert fh.ntrack == 32

    def test_header(self):
        with open(SAMPLE_32TRACK_FANOUT2, 'rb') as fh:
            fh.seek(17436)
            header = mark4.Mark4Header.fromfile(fh, ntrack=32, decade=2010)

        # Try initialising with properties instead of keywords.
        # * time imply the decade, bcd_unit_year, bcd_day, bcd_hour,
        #   bcd_minute, bcd_second, bcd_fraction;
        # * ntrack, samples_per_frame, bps define headstack_id, bcd_track_id,
        #   fan_out, and magnitude_bit;
        # * header.converter since lsb_output and converter_id are somewhat
        #   non-standard
        header1 = mark4.Mark4Header.fromvalues(
            ntrack=32, samples_per_frame=40000, bps=2, time=header.time,
            system_id=108, converters=header.converters)
        assert header1 == header

    def test_file_streamer(self, tmpdir):
        with mark4.open(SAMPLE_32TRACK_FANOUT2, 'rs', sample_rate=16*u.MHz,
                        ntrack=32, decade=2010) as fh:
            header0 = fh.header0
            assert fh.samples_per_frame == 40000
            assert fh.sample_rate == 16 * u.MHz
            start_time = fh.start_time
            assert start_time.yday == '2017:063:04:42:26.02500'
            record = fh.read(80000)
            fh_raw_tell1 = fh.fh_raw.tell()
            assert fh_raw_tell1 == 160000 + 17436

        assert np.all(record[:320] == 0.)
        # Compare with: m5d vlbi_b1133/gk049c_ar_no0011.m5a MKIV1_2-128-8-2 700
        assert np.all(record[320:324].astype(int) == np.array(
            [[-1, -1, 3, 1, 3, 3, 1, 1],
             [-3, -3, 1, -1, -1, 3, -3, -1],
             [-1, -1, -3, -1, 1, 1, -1, 1],
             [-1, -3, -1, 1, -1, 1, -1, 1]]))

        fl = str(tmpdir.join('test.m4'))
        with mark4.open(fl, 'ws', header0=header0, sample_rate=16*u.MHz) as fw:
            fw.write(record)
            number_of_bytes = fw.fh_raw.tell()
            assert number_of_bytes == fh_raw_tell1 - 17436

        # Note: this test would not work if we wrote only a single record.
        with mark4.open(fl, 'rs', ntrack=32, decade=2010,
                        sample_rate=16*u.MHz) as fh:
            assert fh.start_time == start_time
            record2 = fh.read(1000)
            assert np.all(record2 == record[:1000])

        with open(fl, 'rb') as fh, open(SAMPLE_32TRACK_FANOUT2, 'rb') as fr:
            fr.seek(17436)
            orig_bytes = fr.read(number_of_bytes)
            conv_bytes = fh.read()
            assert conv_bytes == orig_bytes


class Test16TrackFanout4():
    def test_locate_frame(self):
        with mark4.open(SAMPLE_16TRACK, 'rb') as fh:
            assert fh.locate_frame() == 22124
            assert fh.ntrack == 16

    def test_header(self):
        with open(SAMPLE_16TRACK, 'rb') as fh:
            fh.seek(22124)
            header = mark4.Mark4Header.fromfile(fh, ntrack=16, decade=2010)

        # Try initialising with properties instead of keywords.
        # * time imply the decade, bcd_unit_year, bcd_day, bcd_hour,
        #   bcd_minute, bcd_second, bcd_fraction;
        # * ntrack, samples_per_frame, bps define headstack_id, bcd_track_id,
        #   fan_out, and magnitude_bit;
        # * nsb = 1 sets lsb_output and converter_id
        header1 = mark4.Mark4Header.fromvalues(
            ntrack=16, samples_per_frame=80000, bps=2, time=header.time,
            system_id=108, nsb=1)
        assert header1 == header

    def test_file_streamer(self, tmpdir):
        with mark4.open(SAMPLE_16TRACK, 'rs', sample_rate=32*u.MHz,
                        ntrack=16, decade=2010) as fh:
            header0 = fh.header0
            assert fh.samples_per_frame == 80000
            assert fh.sample_rate == 32 * u.MHz
            start_time = fh.start_time
            assert start_time.yday == '2013:307:06:00:00.77000'
            record = fh.read(160000)
            fh_raw_tell1 = fh.fh_raw.tell()
            assert fh_raw_tell1 == 80000 + 22124
            fh.fh_raw.seek(0)
            preheader_junk = fh.fh_raw.read(22124)

        assert np.all(record[:640] == 0.)
        # Compare with m5d ar/gs033a_ar_no0055.m5a MKIV1_4-128-2-2 1000, and
        # taking first 28 payload samples:
        m5access_data = np.array(
            [[3, -3, -1, 1, 1, 1, 1, -1, -3, 3, 3, -1, -1, 3,
              -1, -1, 3, -3, 1, -3, -3, -1, 3, -3, -3, -3, 3, 1],
             [1, 1, -3, -3, 3, 1, -1, 1, 3, 1, 1, 3, -3, -1,
              -1, 1, 1, -3, -1, -1, -3, -3, 1, 3, 1, -1, 1, 3]])
        assert np.all(record[640:668].astype(int) == m5access_data.T)

        fl = str(tmpdir.join('test.m4'))
        with mark4.open(fl, 'ws', header0=header0, sample_rate=32*u.MHz) as fw:
            fw.fh_raw.write(preheader_junk)
            fw.write(record)
            number_of_bytes = fw.fh_raw.tell()
            assert number_of_bytes == fh_raw_tell1

        # Note: this test would not work if we wrote only a single record.
        with mark4.open(fl, 'rs', sample_rate=32*u.MHz,
                        ntrack=16, decade=2010) as fh:
            assert fh.start_time == start_time
            record2 = fh.read()
            assert np.all(record2 == record)

        with open(fl, 'rb') as fh, open(SAMPLE_16TRACK, 'rb') as fr:
            orig_bytes = fr.read(number_of_bytes)
            conv_bytes = fh.read()
            assert conv_bytes == orig_bytes
